from typing import List

import django.db.models.deletion
from django.db import migrations
from django.db import models

import django_afip.models


class Migration(migrations.Migration):

    replaces = [
        ("afip", "0001_initial"),
        ("afip", "0002_auto_20150909_1837"),
        ("afip", "0003_vat_tax_digits"),
        ("afip", "0004_auto_20150916_1934"),
        ("afip", "0005_auto_20150918_0115"),
        ("afip", "0006_auto_20151212_1431"),
        ("afip", "0007_auto_20151212_1754"),
        ("afip", "0008_auto_20151212_1820"),
        ("afip", "0009_auto_20151214_1836"),
        ("afip", "0010_receiptpdf_pdffile_uploadto"),
        ("afip", "0011_receipt_entry_vat"),
        ("afip", "0012_taxpayer_profile_one_to_one"),
        ("afip", "0013_taxpayer_is_sandboxed"),
        ("afip", "0014_no_partially_validated_receiptvalidations"),
        ("afip", "0015_receiptentry__amount_to_quantity"),
        ("afip", "0016_auto_20170529_2012"),
        ("afip", "0017_receipt_issued_date"),
        ("afip", "0018_on_delete_fks"),
        ("afip", "0019_receiptvalidation__processed_date"),
        ("afip", "0020_backfill_receiptvalidation__processed_date"),
        ("afip", "0021_drop_batches"),
        ("afip", "0022_auto_misc_tweaks"),
        ("afip", "0023_taxpayer__certs_blank"),
        ("afip", "0024_taxpayer__certificate_expiration"),
        ("afip", "0025_receipt__default_currency"),
        ("afip", "0026_vat_conditions"),
        ("afip", "0027_taxpayer__active_since"),
        ("afip", "0028_taxpayer__copy_active_since"),
        ("afip", "0029_drop__taxpayerprofile__active_since"),
        ("afip", "0030_receiptpdf_client_vat_condition"),
        ("afip", "0031_receiptpdf__copy_vat_condition"),
        ("afip", "0032_receiptpdf__client_vat__notnull"),
        ("afip", "0033_receiptpdf__pdf_file__help_text"),
        ("afip", "0034_vat_condition_choices"),
        ("afip", "0035_receiptentry__vat__blankable"),
        ("afip", "0036_receiptpdf__client_address__blank"),
    ]

    initial = True

    dependencies: List[str] = []

    operations = [
        migrations.CreateModel(
            name="AuthTicket",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "unique_id",
                    models.IntegerField(
                        default=django_afip.models.default_unique_id,
                        verbose_name="unique id",
                    ),
                ),
                (
                    "generated",
                    models.DateTimeField(
                        default=django_afip.models.default_generated,
                        verbose_name="generated",
                    ),
                ),
                (
                    "expires",
                    models.DateTimeField(
                        default=django_afip.models.default_expires,
                        verbose_name="expires",
                    ),
                ),
                (
                    "service",
                    models.CharField(
                        help_text="Service for which this ticket has been authorized",
                        max_length=6,
                        verbose_name="service",
                    ),
                ),
                ("token", models.TextField(verbose_name="token")),
                ("signature", models.TextField(verbose_name="signature")),
            ],
            options={
                "verbose_name": "authorization ticket",
                "verbose_name_plural": "authorization tickets",
            },
        ),
        migrations.CreateModel(
            name="ConceptType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=3, verbose_name="code")),
                (
                    "description",
                    models.CharField(max_length=250, verbose_name="description"),
                ),
                (
                    "valid_from",
                    models.DateField(blank=True, null=True, verbose_name="valid from"),
                ),
                (
                    "valid_to",
                    models.DateField(blank=True, null=True, verbose_name="valid until"),
                ),
            ],
            options={
                "verbose_name": "concept type",
                "verbose_name_plural": "concept types",
            },
        ),
        migrations.CreateModel(
            name="CurrencyType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=3, verbose_name="code")),
                (
                    "description",
                    models.CharField(max_length=250, verbose_name="description"),
                ),
                (
                    "valid_from",
                    models.DateField(blank=True, null=True, verbose_name="valid from"),
                ),
                (
                    "valid_to",
                    models.DateField(blank=True, null=True, verbose_name="valid until"),
                ),
            ],
            options={
                "verbose_name": "currency type",
                "verbose_name_plural": "currency types",
            },
        ),
        migrations.CreateModel(
            name="DocumentType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=3, verbose_name="code")),
                (
                    "description",
                    models.CharField(max_length=250, verbose_name="description"),
                ),
                (
                    "valid_from",
                    models.DateField(blank=True, null=True, verbose_name="valid from"),
                ),
                (
                    "valid_to",
                    models.DateField(blank=True, null=True, verbose_name="valid until"),
                ),
            ],
            options={
                "verbose_name": "document type",
                "verbose_name_plural": "document types",
            },
        ),
        migrations.CreateModel(
            name="Observation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.PositiveSmallIntegerField(verbose_name="code")),
                ("message", models.CharField(max_length=255, verbose_name="message")),
            ],
            options={
                "verbose_name": "observation",
                "verbose_name_plural": "observations",
            },
        ),
        migrations.CreateModel(
            name="PointOfSales",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("number", models.PositiveSmallIntegerField(verbose_name="number")),
                (
                    "issuance_type",
                    models.CharField(
                        help_text="Indicates if thie POS emits using CAE and CAEA.",
                        max_length=8,
                        verbose_name="issuance type",
                    ),
                ),
                ("blocked", models.BooleanField(verbose_name="blocked")),
                (
                    "drop_date",
                    models.DateField(blank=True, null=True, verbose_name="drop date"),
                ),
            ],
            options={
                "verbose_name": "point of sales",
                "verbose_name_plural": "points of sales",
            },
        ),
        migrations.CreateModel(
            name="Receipt",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "document_number",
                    models.BigIntegerField(
                        help_text=(
                            "The document number of the customer to whom this receipt"
                            " is addressed"
                        ),
                        verbose_name="document number",
                    ),
                ),
                (
                    "receipt_number",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text=(
                            "If left blank, the next valid number will assigned when"
                            " validating the receipt."
                        ),
                        null=True,
                        verbose_name="receipt number",
                    ),
                ),
                (
                    "issued_date",
                    models.DateField(
                        help_text=(
                            "Can diverge up to 5 days for good, or 10 days otherwise"
                        ),
                        verbose_name="issued date",
                    ),
                ),
                (
                    "total_amount",
                    models.DecimalField(
                        decimal_places=2,
                        help_text=(
                            "Must be equal to the sum of net_taxed, exempt_amount,"
                            " net_taxes, and all taxes and vats."
                        ),
                        max_digits=15,
                        verbose_name="total amount",
                    ),
                ),
                (
                    "net_untaxed",
                    models.DecimalField(
                        decimal_places=2,
                        help_text=(
                            "The total amount to which taxes do not apply. For C-type"
                            " receipts, this must be zero."
                        ),
                        max_digits=15,
                        verbose_name="total untaxable amount",
                    ),
                ),
                (
                    "net_taxed",
                    models.DecimalField(
                        decimal_places=2,
                        help_text=(
                            "The total amount to which taxes apply. For C-type"
                            " receipts, this is equal to the subtotal."
                        ),
                        max_digits=15,
                        verbose_name="total taxable amount",
                    ),
                ),
                (
                    "exempt_amount",
                    models.DecimalField(
                        decimal_places=2,
                        help_text=(
                            "Only for categories which are tax-exempt. For C-type"
                            " receipts, this must be zero."
                        ),
                        max_digits=15,
                        verbose_name="exempt amount",
                    ),
                ),
                (
                    "service_start",
                    models.DateField(
                        blank=True,
                        help_text=(
                            "Date on which a service started. No applicable for goods."
                        ),
                        null=True,
                        verbose_name="service start date",
                    ),
                ),
                (
                    "service_end",
                    models.DateField(
                        blank=True,
                        help_text=(
                            "Date on which a service ended. No applicable for goods."
                        ),
                        null=True,
                        verbose_name="service end date",
                    ),
                ),
                (
                    "expiration_date",
                    models.DateField(
                        blank=True,
                        help_text=(
                            "Date on which this receipt expires. No applicable for"
                            " goods."
                        ),
                        null=True,
                        verbose_name="receipt expiration date",
                    ),
                ),
                (
                    "currency_quote",
                    models.DecimalField(
                        decimal_places=6,
                        default=1,
                        help_text=(
                            "Quote of the day for the currency used in the receipt"
                        ),
                        max_digits=10,
                        verbose_name="currency quote",
                    ),
                ),
                (
                    "concept",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="receipts",
                        to="afip.ConceptType",
                        verbose_name="concept",
                    ),
                ),
                (
                    "currency",
                    models.ForeignKey(
                        default=django_afip.models.first_currency,
                        help_text="Currency in which this receipt is issued.",
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="documents",
                        to="afip.CurrencyType",
                        verbose_name="currency",
                    ),
                ),
                (
                    "document_type",
                    models.ForeignKey(
                        help_text=(
                            "The document type of the customer to whom this receipt is"
                            " addressed"
                        ),
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="receipts",
                        to="afip.DocumentType",
                        verbose_name="document type",
                    ),
                ),
                (
                    "point_of_sales",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="receipts",
                        to="afip.PointOfSales",
                        verbose_name="point of sales",
                    ),
                ),
            ],
            options={
                "verbose_name": "receipt",
                "verbose_name_plural": "receipts",
                "ordering": ("issued_date",),
            },
        ),
        migrations.CreateModel(
            name="ReceiptEntry",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "description",
                    models.CharField(max_length=128, verbose_name="description"),
                ),
                ("quantity", models.PositiveSmallIntegerField(verbose_name="quantity")),
                (
                    "unit_price",
                    models.DecimalField(
                        decimal_places=2,
                        help_text="Price per unit before vat or taxes.",
                        max_digits=15,
                        verbose_name="unit price",
                    ),
                ),
                (
                    "receipt",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="entries",
                        to="afip.Receipt",
                        verbose_name="receipt",
                    ),
                ),
            ],
            options={
                "verbose_name": "receipt entry",
                "verbose_name_plural": "receipt entries",
            },
        ),
        migrations.CreateModel(
            name="ReceiptPDF",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "pdf_file",
                    models.FileField(
                        blank=True,
                        help_text="The actual file which contains the PDF data.",
                        null=True,
                        upload_to="receipts",
                        verbose_name="pdf file",
                    ),
                ),
                (
                    "issuing_name",
                    models.CharField(max_length=128, verbose_name="issuing name"),
                ),
                ("issuing_address", models.TextField(verbose_name="issuing address")),
                (
                    "issuing_email",
                    models.CharField(
                        blank=True,
                        max_length=128,
                        null=True,
                        verbose_name="issuing email",
                    ),
                ),
                (
                    "vat_condition",
                    models.CharField(
                        choices=[
                            ("IVA Responsable Inscripto", "IVA Responsable Inscripto"),
                            (
                                "IVA Responsable No Inscripto",
                                "IVA Responsable No Inscripto",
                            ),
                            ("IVA Exento", "IVA Exento"),
                            ("No Responsable IVA", "No Responsable IVA"),
                            ("Responsable Monotributo", "Responsable Monotributo"),
                        ],
                        max_length=48,
                        verbose_name="vat condition",
                    ),
                ),
                (
                    "gross_income_condition",
                    models.CharField(
                        max_length=48, verbose_name="gross income condition"
                    ),
                ),
                (
                    "client_name",
                    models.CharField(max_length=128, verbose_name="client name"),
                ),
                (
                    "client_address",
                    models.TextField(blank=True, verbose_name="client address"),
                ),
                (
                    "client_vat_condition",
                    models.CharField(
                        choices=[
                            ("IVA Responsable Inscripto", "IVA Responsable Inscripto"),
                            (
                                "IVA Responsable No Inscripto",
                                "IVA Responsable No Inscripto",
                            ),
                            ("IVA Sujeto Exento", "IVA Sujeto Exento"),
                            ("Consumidor Final", "Consumidor Final"),
                            ("Responsable Monotributo", "Responsable Monotributo"),
                            ("Proveedor del Exterior", "Proveedor del Exterior"),
                            ("Cliente del Exterior", "Cliente del Exterior"),
                            (
                                "IVA Liberado - Ley Nº 19.640",
                                "IVA Liberado - Ley Nº 19.640",
                            ),
                            (
                                "IVA Responsable Inscripto - Agente de Percepción",
                                "IVA Responsable Inscripto - Agente de Percepción",
                            ),
                            ("Monotributista Social", "Monotributista Social"),
                            ("IVA no alcanzado", "IVA no alcanzado"),
                        ],
                        max_length=48,
                        verbose_name="client vat condition",
                    ),
                ),
                (
                    "sales_terms",
                    models.CharField(
                        help_text=(
                            'Should be something like "Cash", "Payable in 30 days", etc'
                        ),
                        max_length=48,
                        verbose_name="sales terms",
                    ),
                ),
                (
                    "receipt",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.PROTECT,
                        to="afip.Receipt",
                        verbose_name="receipt",
                    ),
                ),
            ],
            options={
                "verbose_name": "receipt pdf",
                "verbose_name_plural": "receipt pdfs",
            },
        ),
        migrations.CreateModel(
            name="ReceiptType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=3, verbose_name="code")),
                (
                    "description",
                    models.CharField(max_length=250, verbose_name="description"),
                ),
                (
                    "valid_from",
                    models.DateField(blank=True, null=True, verbose_name="valid from"),
                ),
                (
                    "valid_to",
                    models.DateField(blank=True, null=True, verbose_name="valid until"),
                ),
            ],
            options={
                "verbose_name": "receipt type",
                "verbose_name_plural": "receipt types",
            },
        ),
        migrations.CreateModel(
            name="ReceiptValidation",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "result",
                    models.CharField(
                        choices=[("A", "approved"), ("R", "rejected")],
                        help_text=(
                            "Indicates whether the validation was succesful or not"
                        ),
                        max_length=1,
                        verbose_name="result",
                    ),
                ),
                ("processed_date", models.DateTimeField(verbose_name="processed date")),
                (
                    "cae",
                    models.CharField(
                        help_text="The CAE as returned by the AFIP",
                        max_length=14,
                        verbose_name="cae",
                    ),
                ),
                (
                    "cae_expiration",
                    models.DateField(
                        help_text="The CAE expiration as returned by the AFIP",
                        verbose_name="cae expiration",
                    ),
                ),
                (
                    "observations",
                    models.ManyToManyField(
                        help_text=(
                            "The observations as returned by the AFIP. These are"
                            " generally present for failed validations."
                        ),
                        related_name="validations",
                        to="afip.Observation",
                        verbose_name="observations",
                    ),
                ),
                (
                    "receipt",
                    models.OneToOneField(
                        help_text="The Receipt for which this validation applies",
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="validation",
                        to="afip.Receipt",
                        verbose_name="receipt",
                    ),
                ),
            ],
            options={
                "verbose_name": "receipt validation",
                "verbose_name_plural": "receipt validations",
            },
        ),
        migrations.CreateModel(
            name="Tax",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "description",
                    models.CharField(max_length=80, verbose_name="description"),
                ),
                (
                    "base_amount",
                    models.DecimalField(
                        decimal_places=2, max_digits=15, verbose_name="base amount"
                    ),
                ),
                (
                    "aliquot",
                    models.DecimalField(
                        decimal_places=2, max_digits=5, verbose_name="aliquot"
                    ),
                ),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2, max_digits=15, verbose_name="amount"
                    ),
                ),
                (
                    "receipt",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="taxes",
                        to="afip.Receipt",
                    ),
                ),
            ],
            options={"verbose_name": "tax", "verbose_name_plural": "taxes"},
        ),
        migrations.CreateModel(
            name="TaxPayer",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="A friendly name to recognize this taxpayer.",
                        max_length=32,
                        verbose_name="name",
                    ),
                ),
                (
                    "key",
                    models.FileField(
                        blank=True, null=True, upload_to="", verbose_name="key"
                    ),
                ),
                (
                    "certificate",
                    models.FileField(
                        blank=True, null=True, upload_to="", verbose_name="certificate"
                    ),
                ),
                ("cuit", models.BigIntegerField(verbose_name="cuit")),
                (
                    "is_sandboxed",
                    models.BooleanField(
                        help_text=(
                            "Indicates if this taxpayer interacts with the sandbox"
                            " servers rather than the production servers"
                        ),
                        verbose_name="is sandboxed",
                    ),
                ),
                (
                    "certificate_expiration",
                    models.DateTimeField(
                        editable=False,
                        help_text=(
                            "Stores expiration for the current certificate. Note that"
                            " this field is updated pre-save, so the value may be"
                            " invalid for unsaved models."
                        ),
                        null=True,
                        verbose_name="certificate expiration",
                    ),
                ),
                (
                    "active_since",
                    models.DateField(
                        help_text=(
                            "Date since which this taxpayer has been legally active."
                        ),
                        verbose_name="active since",
                    ),
                ),
            ],
            options={"verbose_name": "taxpayer", "verbose_name_plural": "taxpayers"},
        ),
        migrations.CreateModel(
            name="TaxPayerProfile",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "issuing_name",
                    models.CharField(max_length=128, verbose_name="issuing name"),
                ),
                ("issuing_address", models.TextField(verbose_name="issuing address")),
                (
                    "issuing_email",
                    models.CharField(
                        blank=True,
                        max_length=128,
                        null=True,
                        verbose_name="issuing email",
                    ),
                ),
                (
                    "vat_condition",
                    models.CharField(
                        choices=[
                            ("IVA Responsable Inscripto", "IVA Responsable Inscripto"),
                            (
                                "IVA Responsable No Inscripto",
                                "IVA Responsable No Inscripto",
                            ),
                            ("IVA Exento", "IVA Exento"),
                            ("No Responsable IVA", "No Responsable IVA"),
                            ("Responsable Monotributo", "Responsable Monotributo"),
                        ],
                        max_length=48,
                        verbose_name="vat condition",
                    ),
                ),
                (
                    "gross_income_condition",
                    models.CharField(
                        max_length=48, verbose_name="gross income condition"
                    ),
                ),
                (
                    "sales_terms",
                    models.CharField(
                        help_text=(
                            "The terms of the sale printed onto receipts by default"
                            " (eg: single payment, checking account, etc)."
                        ),
                        max_length=48,
                        verbose_name="sales terms",
                    ),
                ),
                (
                    "taxpayer",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="profile",
                        to="afip.TaxPayer",
                        verbose_name="taxpayer",
                    ),
                ),
            ],
            options={
                "verbose_name": "taxpayer profile",
                "verbose_name_plural": "taxpayer profiles",
            },
        ),
        migrations.CreateModel(
            name="TaxType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=3, verbose_name="code")),
                (
                    "description",
                    models.CharField(max_length=250, verbose_name="description"),
                ),
                (
                    "valid_from",
                    models.DateField(blank=True, null=True, verbose_name="valid from"),
                ),
                (
                    "valid_to",
                    models.DateField(blank=True, null=True, verbose_name="valid until"),
                ),
            ],
            options={"verbose_name": "tax type", "verbose_name_plural": "tax types"},
        ),
        migrations.CreateModel(
            name="Vat",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "base_amount",
                    models.DecimalField(
                        decimal_places=2, max_digits=15, verbose_name="base amount"
                    ),
                ),
                (
                    "amount",
                    models.DecimalField(
                        decimal_places=2, max_digits=15, verbose_name="amount"
                    ),
                ),
                (
                    "receipt",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="vat",
                        to="afip.Receipt",
                    ),
                ),
            ],
            options={"verbose_name": "vat", "verbose_name_plural": "vat"},
        ),
        migrations.CreateModel(
            name="VatType",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("code", models.CharField(max_length=3, verbose_name="code")),
                (
                    "description",
                    models.CharField(max_length=250, verbose_name="description"),
                ),
                (
                    "valid_from",
                    models.DateField(blank=True, null=True, verbose_name="valid from"),
                ),
                (
                    "valid_to",
                    models.DateField(blank=True, null=True, verbose_name="valid until"),
                ),
            ],
            options={"verbose_name": "vat type", "verbose_name_plural": "vat types"},
        ),
        migrations.AddField(
            model_name="vat",
            name="vat_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to="afip.VatType",
                verbose_name="vat type",
            ),
        ),
        migrations.AddField(
            model_name="tax",
            name="tax_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                to="afip.TaxType",
                verbose_name="tax type",
            ),
        ),
        migrations.AddField(
            model_name="receiptentry",
            name="vat",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="receipt_entries",
                to="afip.VatType",
                verbose_name="vat",
            ),
        ),
        migrations.AddField(
            model_name="receipt",
            name="receipt_type",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.PROTECT,
                related_name="receipts",
                to="afip.ReceiptType",
                verbose_name="receipt type",
            ),
        ),
        migrations.AddField(
            model_name="receipt",
            name="related_receipts",
            field=models.ManyToManyField(
                blank=True, to="afip.Receipt", verbose_name="related receipts"
            ),
        ),
        migrations.AddField(
            model_name="pointofsales",
            name="owner",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="points_of_sales",
                to="afip.TaxPayer",
                verbose_name="owner",
            ),
        ),
        migrations.AddField(
            model_name="authticket",
            name="owner",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="auth_tickets",
                to="afip.TaxPayer",
                verbose_name="owner",
            ),
        ),
        migrations.AlterUniqueTogether(
            name="receipt",
            unique_together={("point_of_sales", "receipt_type", "receipt_number")},
        ),
        migrations.AlterUniqueTogether(
            name="pointofsales",
            unique_together={("number", "owner")},
        ),
    ]



"""Tests for provided template tags."""
from django_afip.templatetags.django_afip import format_cuit


def test_format_cuit_tag_with_string():
    """Test valid string inputs."""
    assert format_cuit("20329642330") == "20-32964233-0"
    assert format_cuit("20-32964233-0") == "20-32964233-0"


def test_format_cuit_tag_with_number():
    """Test valid numerical input."""
    assert format_cuit(20329642330) == "20-32964233-0"


def test_format_cuit_tag_with_bad_string():
    """Test invalid string input."""
    assert format_cuit("blah blah") == "blah blah"


def test_format_cuit_tag_with_bad_number():
    """Test invalid numerical input."""
    assert format_cuit(1234) == 1234



import pytest
from django.core.management import call_command


@pytest.mark.django_db
def test_no_missing_migrations():
    """Check that there are no missing migrations of the app."""

    # This returns pending migrations -- or false if non are pending.
    assert not call_command("makemigrations", check=True)



import pytest
from django.conf import settings
from django.core import serializers

from django_afip import models
from django_afip.exceptions import AuthenticationError
from django_afip.factories import TaxPayerFactory
from django_afip.factories import get_test_file
from django_afip.models import AuthTicket

CACHED_TICKET_PATH = settings.BASE_DIR / "test_ticket.yaml"
_live_mode = False


def pytest_runtest_setup(item):
    """Set live mode if the marker has been passed to pytest.

    This avoid accidentally using any of the live-mode fixtures in non-live mode."""
    if list(item.iter_markers(name="live")):
        global _live_mode
        _live_mode = True


@pytest.fixture
def expired_crt() -> bytes:
    with open(get_test_file("test_expired.crt"), "rb") as crt:
        return crt.read()


@pytest.fixture
def expired_key() -> bytes:
    with open(get_test_file("test_expired.key"), "rb") as key:
        return key.read()


@pytest.fixture
def live_taxpayer(db):
    """Return a taxpayer usable with AFIP's test servers."""
    return TaxPayerFactory(pk=1)


@pytest.fixture
def live_ticket(db, live_taxpayer):
    """Return an authentication ticket usable with AFIP's test servers.

    AFIP doesn't allow requesting tickets too often, so we after a few runs
    of the test suite, we can't generate tickets any more and have to wait.

    This helper generates a ticket, and saves it to disk into the app's
    BASE_DIR, so that developers can run tests over and over without having to
    worry about the limitation.

    Expired tickets are not deleted or handled properly; it's up to you to
    delete stale cached tickets.

    When running in CI pipelines, this file will never be preset so won't be a
    problem.
    """
    assert _live_mode

    # Try reading a cached ticket from disk:
    try:
        with open(CACHED_TICKET_PATH) as f:
            [obj] = serializers.deserialize("yaml", f.read())
            obj.save()
    except FileNotFoundError:
        # If something failed, we should still have no tickets in the DB:
        assert models.AuthTicket.objects.count() == 0

    try:
        # Get a new ticket. If the one we just loaded is still valid, that one
        # will be returned, otherwise, a new one will be created.
        ticket = AuthTicket.objects.get_any_active("wsfe")
    except AuthenticationError as e:
        pytest.exit(f"Bailing due to failure authenticating with AFIP:\n{e}")

    # No matter how we go it, we must have at least one ticket in the DB:
    assert models.AuthTicket.objects.count() >= 1

    data = serializers.serialize("yaml", [ticket])
    with open(CACHED_TICKET_PATH, "w") as f:
        f.write(data)

    return ticket


@pytest.fixture
def populated_db(live_ticket, live_taxpayer):
    """Populate the database with fixtures and a POS"""

    models.load_metadata()
    live_taxpayer.fetch_points_of_sales()



import base64
import json
import logging
from io import BytesIO

import qrcode
from PIL import Image

from django_afip.models import Receipt

logger = logging.getLogger(__name__)


class ReceiptQrCode:
    # Note: There's a space in the middle in the docs.
    # I'm assuming it's a human error and that the URL doesn't take a space.
    BASE_URL = "https://www.afip.gob.ar/fe/qr/?p="

    def __init__(self, receipt: Receipt):
        self._receipt = receipt
        # The examples on the website say that "importe" and "ctz" are both "decimal"
        # type. JS/JSON has no decimal type. The examples use integeres.
        #
        # Using integers would drop cents, which would result in mismatching
        # information. Using strings would add quotes, which likely breaks.
        #
        # Using floats seems to be the only viable solution, and SHOULD be fine for
        # values in the range supported.
        self._data = {
            "ver": 1,
            "fecha": receipt.issued_date.strftime("%Y-%m-%d"),
            "cuit": str(receipt.point_of_sales.owner.cuit),
            "ptoVta": receipt.point_of_sales.number,
            "tipoCmp": receipt.receipt_type.code,
            "nroCmp": receipt.receipt_number,
            "importe": float(receipt.total_amount),
            "moneda": receipt.currency.code,
            "ctz": float(receipt.currency_quote),
            "tipoDocRec": receipt.document_type.code,
            "nroDocRec": receipt.document_number,
            "tipoCodAut": "E",  # We don't support anything except CAE.
            "codAut": receipt.validation.cae,
        }

    def as_png(self) -> Image:
        json_data = json.dumps(self._data)
        encoded_json = base64.b64encode(json_data.encode()).decode()
        url = f"{self.BASE_URL}{encoded_json}"

        qr = qrcode.QRCode(version=1, border=4)
        qr.add_data(url)
        qr.make()

        return qr.make_image()


def get_encoded_qrcode(receipt_pdf) -> str:
    """Return a QRCode encoded for embeding in HTML."""

    img_data = BytesIO()
    qr_img = ReceiptQrCode(receipt_pdf.receipt).as_png()
    qr_img.save(img_data, format="PNG")

    return base64.b64encode(img_data.getvalue()).decode()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html
# -- Path setup --------------------------------------------------------------
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
from os.path import abspath
from os.path import dirname
from os.path import join

import django

import django_afip

BASE_DIR = dirname(dirname(abspath(__file__)))

sys.path.insert(0, abspath(join(dirname(__file__), "_ext")))
sys.path.insert(0, abspath(join(BASE_DIR, "testapp")))

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "testapp.settings")

django.setup()

# -- Project information -----------------------------------------------------

project = "django-afip"
copyright = "2015-2020, Hugo Osvaldo Barrera"
author = "Hugo Osvaldo Barrera"

# The short X.Y version.
version = django_afip.__version__
# The full version, including alpha/beta/rc tags
release = django_afip.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "django_models",
    "sphinx.ext.intersphinx",
]

intersphinx_mapping = {
    "django": (
        "https://docs.djangoproject.com/en/stable/",
        "https://docs.djangoproject.com/en/stable/_objects/",
    ),
    "weasyprint": (
        "https://doc.courtbouillon.org/weasyprint/stable/",
        None,
    ),
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]


# The master toctree document.
master_doc = "index"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]



from datetime import datetime

import pytest
from factory.django import FileField
from freezegun import freeze_time
from OpenSSL import crypto

from django_afip import factories


@pytest.mark.django_db
def test_key_generation():
    taxpayer = factories.TaxPayerFactory.build(key=None)
    taxpayer.generate_key()

    key = taxpayer.key.file.read().decode()
    assert key.splitlines()[0] == "-----BEGIN PRIVATE KEY-----"
    assert key.splitlines()[-1] == "-----END PRIVATE KEY-----"

    loaded_key = crypto.load_privatekey(crypto.FILETYPE_PEM, key)
    assert isinstance(loaded_key, crypto.PKey)


def test_dont_overwrite_keys():
    text = b"Hello! I'm not really a key :D"
    taxpayer = factories.TaxPayerFactory.build(key=FileField(data=text))

    taxpayer.generate_key()
    key = taxpayer.key.read()

    assert text == key


@pytest.mark.django_db
def test_overwrite_keys_force():
    text = b"Hello! I'm not really a key :D"
    taxpayer = factories.TaxPayerFactory.build(key__data=text)

    taxpayer.generate_key(force=True)
    key = taxpayer.key.file.read().decode()

    assert text != key
    assert key.splitlines()[0] == "-----BEGIN PRIVATE KEY-----"
    assert key.splitlines()[-1] == "-----END PRIVATE KEY-----"

    loaded_key = crypto.load_privatekey(crypto.FILETYPE_PEM, key)
    assert isinstance(loaded_key, crypto.PKey)


@freeze_time(datetime.fromtimestamp(1489537017))
@pytest.mark.django_db
def test_csr_generation():
    taxpayer = factories.TaxPayerFactory.build(key=None)
    taxpayer.generate_key()

    csr_file = taxpayer.generate_csr()
    csr = csr_file.read().decode()

    assert csr.splitlines()[0] == "-----BEGIN CERTIFICATE REQUEST-----"

    assert csr.splitlines()[-1] == "-----END CERTIFICATE REQUEST-----"

    loaded_csr = crypto.load_certificate_request(crypto.FILETYPE_PEM, csr)
    assert isinstance(loaded_csr, crypto.X509Req)

    expected_components = [
        (b"O", b"John Smith"),
        (b"CN", b"djangoafip1489537017"),
        (b"serialNumber", b"CUIT 20329642330"),
    ]

    assert expected_components == loaded_csr.get_subject().get_components()


def test_certificate_object():
    taxpayer = factories.TaxPayerFactory.build()
    cert = taxpayer.certificate_object

    assert isinstance(cert, crypto.X509)


def test_null_certificate_object():
    taxpayer = factories.TaxPayerFactory.build(certificate=None)
    cert = taxpayer.certificate_object

    assert cert is None


def test_expiration_getter():
    taxpayer = factories.TaxPayerFactory.build(certificate=None)
    expiration = taxpayer.get_certificate_expiration()

    assert expiration is None


def test_expiration_getter_no_cert():
    taxpayer = factories.TaxPayerFactory.build()
    expiration = taxpayer.get_certificate_expiration()

    assert isinstance(expiration, datetime)


@pytest.mark.django_db
def test_expiration_signal_update():
    taxpayer = factories.TaxPayerFactory(certificate_expiration=None)
    taxpayer.save()
    expiration = taxpayer.certificate_expiration

    assert isinstance(expiration, datetime)



from datetime import date
from datetime import datetime

import pytz

from django_afip import parsers


def test_parse_null_datetime():
    assert parsers.parse_datetime("NULL") is None


def test_parse_none_datetime():
    assert parsers.parse_datetime(None) is None


def test_parse_datetimes():
    tz = pytz.timezone(pytz.country_timezones["ar"][0])

    assert parsers.parse_datetime("20170730154330") == datetime(
        2017, 7, 30, 15, 43, 30, tzinfo=tz
    )


def test_parse_null_date():
    assert parsers.parse_date("NULL") is None


def test_parse_none_date():
    assert parsers.parse_date(None) is None


def test_parse_dates():
    assert parsers.parse_date("20170730") == date(2017, 7, 30)


def test_weirdly_encoded():
    # This is the encoding AFIP sometimes uses:
    string = "AÃ±adir paÃ\xads"
    assert parsers.parse_string(string) == "Añadir país"



from unittest.mock import patch

import pytest

from django_afip.clients import get_client


def test_services_are_cached():
    service1 = get_client("wsfe", False)
    with patch.dict("django_afip.clients.WSDLS", values={}, clear=True):
        service2 = get_client("wsfe", False)

    assert service1 is service2


def test_inexisting_service():
    with pytest.raises(ValueError, match="Unknown service name, nonexistant"):
        get_client("nonexistant", False)



"""Script used to generate AFIP metadata fixtures.

This script is used to generate the fixtures with AFIP metadata. It is only used to
BUILD django_afip, and should not be used directly by third-party apps.
"""
import django
from django.core import management

if __name__ == "__main__":
    # Set up django...
    django.setup()
    management.call_command("migrate")

    from django_afip.factories import TaxPayerFactory
    from django_afip.models import GenericAfipType

    # Initialise (uses test credentials).
    TaxPayerFactory()

    # Fetch and dump data:
    for model in GenericAfipType.SUBCLASSES:
        model.objects.populate()

        label = model._meta.label.split(".")[1].lower()

        management.call_command(
            "dumpdata",
            f"afip.{label}",
            format="yaml",
            use_natural_primary_keys=True,
            output=f"django_afip/fixtures/{label}.yaml",
        )



import pytest

from django_afip.helpers import ServerStatus
from django_afip.helpers import get_server_status


@pytest.mark.live
def test_get_server_status_production():
    status = get_server_status(True)

    assert isinstance(status, ServerStatus)


@pytest.mark.live
def test_get_server_status_testing():
    status = get_server_status(False)

    assert isinstance(status, ServerStatus)


@pytest.mark.live
def test_server_status_is_true():
    server_status = ServerStatus(app=True, db=True, auth=True)

    assert server_status


@pytest.mark.live
def test_server_status_is_false():
    server_status = ServerStatus(app=False, db=False, auth=False)

    assert not server_status



from __future__ import annotations

from datetime import date
from datetime import datetime

import pytz

TZ_AR = pytz.timezone(pytz.country_timezones["ar"][0])


def parse_datetime(datestring: str | None) -> datetime | None:
    if datestring == "NULL" or datestring is None:
        return None
    return datetime.strptime(datestring, "%Y%m%d%H%M%S").replace(tzinfo=TZ_AR)


def parse_date(datestring: str | None) -> date | None:
    if datestring == "NULL" or datestring is None:
        return None
    return datetime.strptime(datestring, "%Y%m%d").date()


def parse_string(string: str) -> str:
    """Re-encodes strings from AFIP's weird encoding to UTF-8."""
    try:
        return string.encode("latin-1").decode()
    except UnicodeDecodeError:
        # It looks like SOME errors are plain UTF-8 text.
        return string



from __future__ import annotations

import base64
import logging
import os
import random
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from io import BytesIO
from tempfile import NamedTemporaryFile
from typing import BinaryIO
from uuid import uuid4

import pytz
from django.conf import settings
from django.core import management
from django.core.files import File
from django.core.files.storage import Storage
from django.db import models
from django.db.models import Count
from django.db.models import Sum
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from django_renderpdf.helpers import render_pdf
from lxml import etree
from lxml.builder import E
from OpenSSL.crypto import FILETYPE_PEM
from OpenSSL.crypto import X509
from OpenSSL.crypto import load_certificate
from zeep.exceptions import Fault

from . import clients
from . import crypto
from . import exceptions
from . import parsers
from . import serializers

logger = logging.getLogger(__name__)
TZ_AR = pytz.timezone(pytz.country_timezones["ar"][0])

# http://www.afip.gov.ar/afip/resol1415_anexo2.html
VAT_CONDITIONS = (
    "IVA Responsable Inscripto",
    "IVA Responsable No Inscripto",
    "IVA Exento",
    "No Responsable IVA",
    "Responsable Monotributo",
)
# NOTE: If you find a VAT condition not listed here, please open an issue, and
# a reference to where it's defined.
CLIENT_VAT_CONDITIONS = (
    "IVA Responsable Inscripto",
    "IVA Responsable No Inscripto",
    "IVA Sujeto Exento",
    "Consumidor Final",
    "Responsable Monotributo",
    "Proveedor del Exterior",
    "Cliente del Exterior",
    "IVA Liberado - Ley Nº 19.640",
    "IVA Responsable Inscripto - Agente de Percepción",
    "Monotributista Social",
    "IVA no alcanzado",
)


def load_metadata() -> None:
    """Loads metadata from fixtures into the database."""

    for model in GenericAfipType.SUBCLASSES:
        label = model._meta.label.split(".")[1].lower()
        management.call_command("loaddata", label, app="afip")


def check_response(response) -> None:
    """Check that a response is not an error.

    AFIP allows us to create valid tickets with invalid key/CUIT pairs, so we
    can end up with tickets that fail on any service.

    Due to how zeep works, we can't quite catch these sort of errors on some
    middleware level (well, honestly, we need to do a large refactor).

    This method checks if responses have an error, and raise a readable
    message.
    """
    if response.Errors:
        raise exceptions.AfipException(response)


def first_currency() -> int | None:
    """Returns the id for the first currency

    The `default` parameter of a foreign key *MUST* be a primary key (and not
    an instance), else migrations break. This helper method exists solely for
    that purpose.
    """
    ct = CurrencyType.objects.filter(code="PES").first()
    if ct:
        return ct.pk


def _get_storage_from_settings(setting_name: str) -> Storage:
    path = getattr(settings, setting_name, None)
    if not path:
        return import_string(settings.DEFAULT_FILE_STORAGE)()
    return import_string(path)


class GenericAfipTypeManager(models.Manager):
    """Default Manager for GenericAfipType."""

    def __init__(self, service_name: str, type_name: str):
        """Create a new Manager instance for a GenericAfipType.

        This should generally only be required to manually populate a single
        type with upstream data.
        """
        super().__init__()
        self.__service_name = service_name
        self.__type_name = type_name

    def populate(self, ticket: AuthTicket = None) -> None:
        """Fetch and save data for this model from AFIP's WS.

        Direct usage of this method is discouraged, use
        :func:`~.models.load_metadata` instead.
        """
        ticket = ticket or AuthTicket.objects.get_any_active("wsfe")
        client = clients.get_client("wsfe", ticket.owner.is_sandboxed)
        service = getattr(client.service, self.__service_name)
        response_xml = service(serializers.serialize_ticket(ticket))

        check_response(response_xml)

        for result in getattr(response_xml.ResultGet, self.__type_name):
            self.get_or_create(
                code=result.Id,
                description=result.Desc,
                valid_from=parsers.parse_date(result.FchDesde),
                valid_to=parsers.parse_date(result.FchHasta),
            )

    def get_by_natural_key(self, code: str) -> GenericAfipType:
        return self.get(code=code)

    def exists_by_natural_key(self, code: str) -> bool:
        return self.filter(code=code).exists()


class GenericAfipType(models.Model):
    """An abstract class for several of AFIP's metadata types.

    You should not use this class directly, only subclasses of it. You should
    not create subclasses of this model unless you really know what you're doing.
    """

    SUBCLASSES: list[type[models.Model]] = []

    def __init_subclass__(cls, **kwargs):
        """Keeps a registry of known subclasses."""
        super().__init_subclass__(**kwargs)
        GenericAfipType.SUBCLASSES.append(cls)

    code = models.CharField(
        _("code"),
        max_length=3,
    )
    description = models.CharField(
        _("description"),
        max_length=250,
    )
    valid_from = models.DateField(
        _("valid from"),
        null=True,
        blank=True,
    )
    valid_to = models.DateField(
        _("valid until"),
        null=True,
        blank=True,
    )

    def natural_key(self) -> tuple[str]:
        return (self.code,)

    def __str__(self) -> str:
        return self.description

    class Meta:
        abstract = True


class ReceiptType(GenericAfipType):
    """An AFIP receipt type.

    See the AFIP's documentation for details on each receipt type.
    """

    objects = GenericAfipTypeManager("FEParamGetTiposCbte", "CbteTipo")

    class Meta:
        verbose_name = _("receipt type")
        verbose_name_plural = _("receipt types")


class ConceptType(GenericAfipType):
    """An AFIP concept type.

    See the AFIP's documentation for details on each concept type.
    """

    objects = GenericAfipTypeManager("FEParamGetTiposConcepto", "ConceptoTipo")

    class Meta:
        verbose_name = _("concept type")
        verbose_name_plural = _("concept types")


class DocumentType(GenericAfipType):
    """An AFIP document type.

    See the AFIP's documentation for details on each document type.
    """

    objects = GenericAfipTypeManager("FEParamGetTiposDoc", "DocTipo")

    class Meta:
        verbose_name = _("document type")
        verbose_name_plural = _("document types")


class VatType(GenericAfipType):
    """An AFIP VAT type.

    See the AFIP's documentation for details on each VAT type.
    """

    objects = GenericAfipTypeManager("FEParamGetTiposIva", "IvaTipo")

    class Meta:
        verbose_name = _("vat type")
        verbose_name_plural = _("vat types")


class TaxType(GenericAfipType):
    """An AFIP tax type.

    See the AFIP's documentation for details on each tax type.
    """

    objects = GenericAfipTypeManager("FEParamGetTiposTributos", "TributoTipo")

    class Meta:
        verbose_name = _("tax type")
        verbose_name_plural = _("tax types")


class CurrencyType(GenericAfipType):
    """An AFIP curreny type.

    See the AFIP's documentation for details on each currency type.
    """

    objects = GenericAfipTypeManager("FEParamGetTiposMonedas", "Moneda")

    def __str__(self) -> str:
        return f"{self.description} ({self.code})"

    class Meta:
        verbose_name = _("currency type")
        verbose_name_plural = _("currency types")


class TaxPayer(models.Model):
    """Represents an AFIP TaxPayer.

    Note that multiple instances of this object can actually represent the same
    taxpayer, each using a different key.

    The following fields are only used for generating printables, and are never
    sent to AFIP, hence, are entirely optional:

    - logo
    """

    name = models.CharField(
        _("name"),
        max_length=32,
        help_text=_("A friendly name to recognize this taxpayer."),
    )
    key = models.FileField(
        verbose_name=_("key"),
        upload_to="afip/taxpayers/keys/",
        storage=_get_storage_from_settings("AFIP_KEY_STORAGE"),
        blank=True,
        null=True,
    )
    certificate = models.FileField(
        verbose_name=_("certificate"),
        upload_to="afip/taxpayers/certs/",
        storage=_get_storage_from_settings("AFIP_CERT_STORAGE"),
        blank=True,
        null=True,
    )
    cuit = models.BigIntegerField(
        _("cuit"),
    )
    is_sandboxed = models.BooleanField(
        _("is sandboxed"),
        help_text=_(
            "Indicates if this taxpayer should use with the sandbox servers "
            "rather than the production servers."
        ),
    )
    certificate_expiration = models.DateTimeField(
        _("certificate expiration"),
        editable=False,
        null=True,  # Either no cert, or and old TaxPayer
        help_text=_(
            "Stores expiration for the current certificate.<br>Note that this "
            "field is updated pre-save, so the value may be invalid for unsaved models."
        ),
    )
    active_since = models.DateField(
        _("active since"),
        help_text=_("Date since which this taxpayer has been legally active."),
    )
    logo = models.ImageField(
        verbose_name=_("logo"),
        upload_to="afip/taxpayers/logos/",
        storage=_get_storage_from_settings("AFIP_LOGO_STORAGE"),
        blank=True,
        null=True,
        help_text=_("A logo to use when generating printable receipts."),
    )

    @property
    def logo_as_data_uri(self) -> str:
        """This TaxPayer's logo as a data uri."""
        _, ext = os.path.splitext(self.logo.file.name)
        with self.logo.open() as f:
            data = base64.b64encode(f.read())

        return "data:image/{};base64,{}".format(
            ext[1:], data.decode()  # Remove the leading dot.
        )

    @property
    def certificate_object(self) -> X509 | None:
        """Returns the certificate as an OpenSSL object

        Returns the certificate as an OpenSSL object (rather than as a file
        object).
        """

        if not self.certificate:
            return None
        self.certificate.seek(0)
        return load_certificate(FILETYPE_PEM, self.certificate.read())

    def get_certificate_expiration(self) -> datetime | None:
        """
        Gets the certificate expiration from the certificate


        Gets the certificate expiration from the certificate file. Note that
        this value is stored into ``certificate_expiration`` when an instance
        is saved, so you should generally prefer that method (since this one
        requires reading and parsing the entire certificate).
        """
        cert = self.certificate_object
        if not cert:
            return None
        not_after = cert.get_notAfter()
        if not not_after:
            return None
        datestring = not_after.decode()
        dt = datetime.strptime(datestring, "%Y%m%d%H%M%SZ")
        return dt.replace(tzinfo=timezone.utc)

    def generate_key(self, force=False) -> bool:
        """
        Creates a key file for this TaxPayer

        Creates a key file for this TaxPayer if it does not have one, and
        immediately saves it.

        A new key will not be generated if one is already set, unless the ``force``
        parameter is true. This is to prevent overwriting a potentially in-use key.

        Returns True if and only if a key was created.
        """
        if self.key and not force:
            logger.warning("Tried to generate key for a taxpayer that already had one")
            return False

        with NamedTemporaryFile(suffix=".key") as file_:
            crypto.create_key(file_)
            self.key = File(file_, name=f"{uuid4().hex}.key")
            self.save()

        return True

    def generate_csr(self, basename="djangoafip") -> BinaryIO:
        """
        Creates a CSR for this TaxPayer's key

        Creates a file-like object that contains the CSR which can be used to
        request a new certificate from AFIP.
        """
        csr = BytesIO()
        crypto.create_csr(
            self.key.file,
            self.name,
            f"{basename}{int(datetime.now().timestamp())}",
            f"CUIT {self.cuit}",
            csr,
        )
        csr.seek(0)
        return csr

    def create_ticket(self, service: str) -> AuthTicket:
        """Create an AuthTicket for a given service."""
        ticket = AuthTicket(owner=self, service=service)
        ticket.authorize()
        return ticket

    def get_ticket(self, service: str) -> AuthTicket | None:
        """Return an existing AuthTicket for a given service, if any."""
        return self.auth_tickets.filter(
            expires__gt=datetime.now(timezone.utc),
            service=service,
        ).last()

    def get_or_create_ticket(self, service: str) -> AuthTicket:
        """
        Return or create a new AuthTicket for a given serivce.

        Return an existing ticket for a service if one is available, otherwise,
        create a new one and return that.

        This is generally the preferred method of obtaining tickets for any
        service.
        """
        return self.get_ticket(service) or self.create_ticket(service)

    def fetch_points_of_sales(self, ticket: AuthTicket = None) -> list[PointOfSales]:
        """
        Fetch all point of sales objects.

        Fetch all point of sales from the WS and store (or update) them
        locally.

        Returns a list of tuples with the format (pos, created,).
        """
        ticket = ticket or self.get_or_create_ticket("wsfe")

        client = clients.get_client("wsfe", self.is_sandboxed)
        response = client.service.FEParamGetPtosVenta(
            serializers.serialize_ticket(ticket),
        )
        check_response(response)

        results = []
        for pos_data in response.ResultGet.PtoVenta:
            results.append(
                PointOfSales.objects.update_or_create(
                    number=pos_data.Nro,
                    owner=self,
                    defaults={
                        "issuance_type": pos_data.EmisionTipo,
                        "blocked": pos_data.Bloqueado == "N",
                        "drop_date": parsers.parse_date(pos_data.FchBaja),
                    },
                )
            )

        return results

    def __repr__(self) -> str:
        return "<TaxPayer {}: {}, CUIT {}>".format(
            self.pk,
            self.name,
            self.cuit,
        )

    def __str__(self) -> str:
        return str(self.name)

    class Meta:
        verbose_name = _("taxpayer")
        verbose_name_plural = _("taxpayers")


class PointOfSales(models.Model):
    """
    Represents an existing AFIP point of sale.

    Points of sales need to be created via AFIP's web interface and it is
    recommended that you use :meth:`~.TaxPayer.fetch_points_of_sales` to fetch
    these programatically.

    Note that deleting or altering these models will not affect upstream point
    of sales.

    This model also contains a few fields that are not required or sent to the
    AFIP when validating receipt. They are used *only* for PDF generation.
    Those fields are:

    - issuing_name
    - issuing_address
    - issuing_email
    - vat_condition
    - gross_income_condition
    - sales_terms

    These fields may be ignored when using an external mechanism to generate
    PDF or printable receipts.
    """

    number = models.PositiveSmallIntegerField(
        _("number"),
    )
    issuance_type = models.CharField(
        _("issuance type"),
        max_length=24,
        help_text="Indicates if this POS emits using CAE and CAEA.",
    )
    blocked = models.BooleanField(
        _("blocked"),
    )
    drop_date = models.DateField(
        _("drop date"),
        null=True,
        blank=True,
    )

    owner = models.ForeignKey(
        TaxPayer,
        related_name="points_of_sales",
        verbose_name=_("owner"),
        on_delete=models.CASCADE,
    )

    # The following fields are only used for PDF generation.
    issuing_name = models.CharField(
        max_length=128,
        null=True,
        verbose_name=_("issuing name"),
        help_text=_("The name of the issuing entity as shown on receipts."),
    )
    issuing_address = models.TextField(
        _("issuing address"),
        null=True,
        help_text=_("The address of the issuing entity as shown on receipts."),
    )
    issuing_email = models.CharField(
        max_length=128,
        verbose_name=_("issuing email"),
        blank=True,
        null=True,
        help_text=_("The email of the issuing entity as shown on receipts."),
    )
    vat_condition = models.CharField(
        max_length=48,
        choices=(
            (
                condition,
                condition,
            )
            for condition in VAT_CONDITIONS
        ),
        null=True,
        verbose_name=_("vat condition"),
    )
    gross_income_condition = models.CharField(
        max_length=48,
        null=True,
        verbose_name=_("gross income condition"),
    )
    sales_terms = models.CharField(
        max_length=48,
        null=True,
        verbose_name=_("sales terms"),
        help_text=_(
            "The terms of the sale printed onto receipts by default "
            "(eg: single payment, checking account, etc)."
        ),
    )

    def __str__(self) -> str:
        return str(self.number)

    class Meta:
        unique_together = (("number", "owner"),)
        verbose_name = _("point of sales")
        verbose_name_plural = _("points of sales")


class AuthTicketManager(models.Manager):
    def get_any_active(self, service: str) -> AuthTicket:
        """Return a valid, active ticket for a given service."""
        ticket = AuthTicket.objects.filter(
            token__isnull=False,
            expires__gt=datetime.now(timezone.utc),
            service=service,
        ).first()
        if ticket:
            return ticket

        taxpayer = TaxPayer.objects.order_by("?").first()

        if not taxpayer:
            raise exceptions.AuthenticationError(
                _("There are no taxpayers to generate a ticket."),
            )

        return taxpayer.create_ticket(service)


def default_generated() -> datetime:
    """The default generated date for new tickets."""
    return datetime.now(TZ_AR)


def default_expires() -> datetime:
    """The default expiration date for new tickets."""
    tomorrow = datetime.now(TZ_AR) + timedelta(hours=12)
    return tomorrow


def default_unique_id() -> int:
    """A random unique id for new tickets."""
    return random.randint(0, 2147483647)


class AuthTicket(models.Model):
    """An AFIP Authorization ticket.

    This is a signed ticket used to communicate with AFIP's webservices.

    Applications should not generally have to deal with these tickets
    themselves; most services will find or create one as necessary.
    """

    owner = models.ForeignKey(
        TaxPayer,
        verbose_name=_("owner"),
        related_name="auth_tickets",
        on_delete=models.CASCADE,
    )
    unique_id = models.IntegerField(
        _("unique id"),
        default=default_unique_id,
    )
    generated = models.DateTimeField(
        _("generated"),
        default=default_generated,
    )
    expires = models.DateTimeField(
        _("expires"),
        default=default_expires,
    )
    service = models.CharField(
        _("service"),
        max_length=6,
        help_text=_("Service for which this ticket has been authorized."),
    )

    token = models.TextField(
        _("token"),
    )
    signature = models.TextField(
        _("signature"),
    )

    objects = AuthTicketManager()

    TOKEN_XPATH = "/loginTicketResponse/credentials/token"
    SIGN_XPATH = "/loginTicketResponse/credentials/sign"

    def __create_request_xml(self) -> bytes:
        request_xml = E.loginTicketRequest(
            {"version": "1.0"},
            E.header(
                E.uniqueId(str(self.unique_id)),
                E.generationTime(serializers.serialize_datetime(self.generated)),
                E.expirationTime(serializers.serialize_datetime(self.expires)),
            ),
            E.service(self.service),
        )
        # Hint: tostring returns bytes.
        return etree.tostring(request_xml, pretty_print=True)

    def __sign_request(self, request: bytes) -> bytes:
        with self.owner.certificate.file.open("rb") as f:
            cert = f.read()

        with self.owner.key.file.open("rb") as f:
            key = f.read()

        return crypto.create_embeded_pkcs7_signature(request, cert, key)

    def authorize(self) -> None:
        """Send this ticket to AFIP for authorization."""
        request_xml = self.__create_request_xml()
        signed_request = self.__sign_request(request_xml)
        request = base64.b64encode(signed_request).decode()

        client = clients.get_client("wsaa", self.owner.is_sandboxed)
        try:
            raw_response = client.service.loginCms(request)
        except Fault as e:
            if str(e) == "Certificado expirado":
                raise exceptions.CertificateExpired(str(e)) from e
            if str(e) == "Certificado no emitido por AC de confianza":
                raise exceptions.UntrustedCertificate(str(e)) from e
            raise exceptions.AuthenticationError(str(e)) from e
        response = etree.fromstring(raw_response.encode("utf-8"))

        self.token = response.xpath(self.TOKEN_XPATH)[0].text
        self.signature = response.xpath(self.SIGN_XPATH)[0].text

        self.save()

    def __str__(self) -> str:
        return str(self.unique_id)

    class Meta:
        verbose_name = _("authorization ticket")
        verbose_name_plural = _("authorization tickets")


class ReceiptQuerySet(models.QuerySet):
    """The default queryset obtains when querying via :class:`~.ReceiptManager`."""

    def _assign_numbers(self) -> None:
        """Assign numbers in preparation for validating these receipts.

        WARNING: Don't call the method manually unless you know what you're
        doing!
        """
        first = self.select_related("point_of_sales", "receipt_type").first()

        next_num = (
            Receipt.objects.fetch_last_receipt_number(
                first.point_of_sales,
                first.receipt_type,
            )
            + 1
        )

        for receipt in self.filter(receipt_number__isnull=True):
            # Atomically update receipt number
            Receipt.objects.filter(pk=receipt.id, receipt_number__isnull=True,).update(
                receipt_number=next_num,
            )
            next_num += 1

    def check_groupable(self) -> ReceiptQuerySet:
        """Check that all receipts returned by this queryset are groupable.

        "Groupable" means that they can be validated together: they have the
        same POS and receipt type.

        Returns the same queryset is all receipts are groupable, otherwise,
        raises :class:`~.CannotValidateTogether`.
        """
        types = self.aggregate(
            poses=Count(
                "point_of_sales_id",
            ),
            types=Count("receipt_type"),
        )

        if set(types.values()) > {1}:
            raise exceptions.CannotValidateTogether()

        return self

    def validate(self, ticket: AuthTicket = None) -> list[str]:
        """Validate all receipts matching this queryset.

        Note that, due to how AFIP implements its numbering, this method is not
        thread-safe, or even multiprocess-safe.

        Because of this, it is possible that not all instances matching this
        queryset are validated properly. Obviously, only successfully validated
        receipts will be updated.

        Returns a list of errors as returned from AFIP's webservices. An
        exception is not raised because partial failures are possible.

        Receipts that succesfully validate will have a
        :class:`~.ReceiptValidation` object attatched to them with a validation
        date and CAE information.

        Already-validated receipts are ignored.

        Attempting to validate an empty queryset will simply return an empty
        list.
        """
        # Skip any already-validated ones:
        qs = self.filter(validation__isnull=True).check_groupable()
        if qs.count() == 0:
            return []
        qs.order_by("issued_date", "id")._assign_numbers()

        return qs._validate(ticket)

    def _validate(self, ticket=None) -> list[str]:
        first = self.first()
        ticket = ticket or first.point_of_sales.owner.get_or_create_ticket("wsfe")
        client = clients.get_client("wsfe", first.point_of_sales.owner.is_sandboxed)
        response = client.service.FECAESolicitar(
            serializers.serialize_ticket(ticket),
            serializers.serialize_multiple_receipts(self),
        )
        check_response(response)
        errs = []
        for cae_data in response.FeDetResp.FECAEDetResponse:
            if cae_data.Resultado == ReceiptValidation.RESULT_APPROVED:
                validation = ReceiptValidation.objects.create(
                    result=cae_data.Resultado,
                    cae=cae_data.CAE,
                    cae_expiration=parsers.parse_date(cae_data.CAEFchVto),
                    receipt=self.get(
                        receipt_number=cae_data.CbteDesde,
                    ),
                    processed_date=parsers.parse_datetime(
                        response.FeCabResp.FchProceso,
                    ),
                )
                if cae_data.Observaciones:
                    for obs in cae_data.Observaciones.Obs:
                        observation = Observation.objects.create(
                            code=obs.Code,
                            message=obs.Msg,
                        )
                    validation.observations.add(observation)
            elif cae_data.Observaciones:
                for obs in cae_data.Observaciones.Obs:
                    errs.append(
                        "Error {}: {}".format(
                            obs.Code,
                            parsers.parse_string(obs.Msg),
                        )
                    )

        # Remove the number from ones that failed to validate:
        self.filter(validation__isnull=True).update(receipt_number=None)

        return errs


class ReceiptManager(models.Manager):
    """Default manager for the :class:`~.Receipt` class.

    You should generally access this using ``Receipt.objects``.
    """

    def fetch_last_receipt_number(
        self,
        point_of_sales: PointOfSales,
        receipt_type: ReceiptType,
    ) -> int:
        """Returns the number for the last validated receipt."""
        client = clients.get_client("wsfe", point_of_sales.owner.is_sandboxed)
        response_xml = client.service.FECompUltimoAutorizado(
            serializers.serialize_ticket(
                point_of_sales.owner.get_or_create_ticket("wsfe")
            ),
            point_of_sales.number,
            receipt_type.code,
        )
        check_response(response_xml)

        # TODO XXX: Error handling
        # (FERecuperaLastCbteResponse){
        #    PtoVta = 0
        #    CbteTipo = 0
        #    CbteNro = 0
        #    Errors =
        #       (ArrayOfErr){
        #          Err[] =
        #             (Err){
        #                Code = 601
        #                Msg = "CUIT representada no incluida en Token"
        #             },
        #       }
        #  }

        return response_xml.CbteNro

    def fetch_receipt_data(
        self,
        receipt_type: ReceiptType,
        receipt_number: int,
        point_of_sales: PointOfSales,
    ):  # TODO: Wrap this in a dataclass
        """Returns receipt related data"""

        if not receipt_number:
            return None

        client = clients.get_client("wsfe", point_of_sales.owner.is_sandboxed)
        response_xml = client.service.FECompConsultar(
            serializers.serialize_ticket(
                point_of_sales.owner.get_or_create_ticket("wsfe")
            ),
            serializers.serialize_receipt_data(
                receipt_type, receipt_number, point_of_sales.number
            ),
        )
        try:
            check_response(response_xml)
            return response_xml.ResultGet
        except exceptions.AfipException:
            return None

    def get_queryset(self) -> ReceiptQuerySet:
        return ReceiptQuerySet(self.model, using=self._db).select_related(
            "receipt_type",
        )


class Receipt(models.Model):
    """A receipt, as sent to AFIP.

    Note that AFIP allows sending ranges of receipts, but this isn't generally
    what you want, so we model invoices individually.

    You'll probably want to relate some `Sale` or `Order` object from your
    model with each Receipt.

    All ``document_`` fields contain the recipient's data.

    If the taxpayer has taxes or pays VAT, you need to attach :class:`~.Tax`
    and/or :class:`~.Vat` instances to the Receipt.
    """

    point_of_sales = models.ForeignKey(
        PointOfSales,
        related_name="receipts",
        verbose_name=_("point of sales"),
        on_delete=models.PROTECT,
    )
    receipt_type = models.ForeignKey(
        ReceiptType,
        related_name="receipts",
        verbose_name=_("receipt type"),
        on_delete=models.PROTECT,
    )
    concept = models.ForeignKey(
        ConceptType,
        verbose_name=_("concept"),
        related_name="receipts",
        on_delete=models.PROTECT,
    )
    document_type = models.ForeignKey(
        DocumentType,
        verbose_name=_("document type"),
        related_name="receipts",
        help_text=_("The document type of the recipient of this receipt."),
        on_delete=models.PROTECT,
    )
    document_number = models.BigIntegerField(
        _("document number"),
        help_text=_("The document number of the recipient of this receipt."),
    )
    # NOTE: WS will expect receipt_from and receipt_to.
    receipt_number = models.PositiveIntegerField(
        _("receipt number"),
        null=True,
        blank=True,
        help_text=_(
            "If left blank, the next valid number will assigned when "
            "validating the receipt."
        ),
    )
    issued_date = models.DateField(
        verbose_name=_("issued date"),
        help_text=_("Can diverge up to 5 days for good, or 10 days otherwise."),
    )
    total_amount = models.DecimalField(
        # ImpTotal
        _("total amount"),
        max_digits=15,
        decimal_places=2,
        help_text=_(
            "Must be equal to the sum of net_taxed, exempt_amount, net_taxes, "
            "and all taxes and vats."
        ),
    )
    net_untaxed = models.DecimalField(
        # ImpTotConc
        _("total untaxable amount"),
        max_digits=15,
        decimal_places=2,
        help_text=_(
            "The total amount to which taxes do not apply.<br>"
            "For C-type receipts, this must be zero."
        ),
    )
    net_taxed = models.DecimalField(
        # ImpNeto
        _("total taxable amount"),
        max_digits=15,
        decimal_places=2,
        help_text=_(
            "The total amount to which taxes apply.<br>"
            "For C-type receipts, this is equal to the subtotal."
        ),
    )
    exempt_amount = models.DecimalField(
        # ImpOpEx
        # Sólo para emisores que son IVA exento
        _("exempt amount"),
        max_digits=15,
        decimal_places=2,
        help_text=_(
            "Only for categories which are tax-exempt.<br>"
            "For C-type receipts, this must be zero."
        ),
    )
    service_start = models.DateField(
        _("service start date"),
        help_text=_("Date on which a service started. No applicable for goods."),
        null=True,
        blank=True,
    )
    service_end = models.DateField(
        _("service end date"),
        help_text=_("Date on which a service ended. No applicable for goods."),
        null=True,
        blank=True,
    )
    expiration_date = models.DateField(
        _("receipt expiration date"),
        help_text=_("Date on which this receipt expires. No applicable for goods."),
        null=True,
        blank=True,
    )
    currency = models.ForeignKey(
        CurrencyType,
        verbose_name=_("currency"),
        related_name="documents",
        help_text=_("Currency in which this receipt is issued."),
        on_delete=models.PROTECT,
        default=first_currency,
    )
    currency_quote = models.DecimalField(
        _("currency quote"),
        max_digits=10,
        decimal_places=6,
        default=1,
        help_text=_("The currency's quote on the day this receipt was issued."),
    )
    related_receipts = models.ManyToManyField(
        "Receipt",
        verbose_name=_("related receipts"),
        blank=True,
    )

    objects = ReceiptManager()

    # TODO: Not implemented: optionals
    # TODO: methods to validate totals

    @property
    def total_vat(self) -> int:
        """Returns the sum of all Vat objects."""
        q = Vat.objects.filter(receipt=self).aggregate(total=Sum("amount"))
        return q["total"] or 0

    @property
    def total_tax(self) -> int:
        """Returns the sum of all Tax objects."""
        q = Tax.objects.filter(receipt=self).aggregate(total=Sum("amount"))
        return q["total"] or 0

    @property
    def formatted_number(self) -> str | None:
        """This receipt's number in the usual format: ``0001-00003087``."""
        if self.receipt_number:
            return "{:04d}-{:08d}".format(
                self.point_of_sales.number,
                self.receipt_number,
            )
        return None

    @property
    def is_validated(self) -> bool:
        """True if this instance is validated.

        Note that resolving this property requires a DB query, so if you've a
        very large amount of receipts you should prefetch (see django's
        ``select_related``) the ``validation`` field. Even so, a DB query *may*
        be triggered.

        If you need a large list of validated receipts, you should actually
        filter them via a QuerySet::

            Receipt.objects.filter(validation__result==RESULT_APPROVED)
        """
        # Avoid the DB lookup if possible:
        if not self.receipt_number:
            return False

        try:
            return self.validation.result == ReceiptValidation.RESULT_APPROVED
        except ReceiptValidation.DoesNotExist:
            return False

    def validate(self, ticket: AuthTicket = None, raise_=False) -> list[str]:
        """Validates this receipt.

        This is a shortcut to :class:`~.ReceiptQuerySet`'s method of the same
        name. Calling this validates only this instance.

        :param AuthTicket ticket: Use this ticket. If None, one will be loaded
            or created automatically.
        :param bool raise_: If True, an exception will be raised when
            validation fails.
        """
        # XXX: Maybe actually have this sortcut raise an exception?
        rv = Receipt.objects.filter(pk=self.pk).validate(ticket)
        # Since we're operating via a queryset, this instance isn't properly
        # updated:
        self.refresh_from_db()
        if raise_ and rv:
            raise exceptions.ValidationError(rv[0])
        return rv

    def revalidate(self) -> ReceiptValidation | None:
        """Revalidate this receipt.

        Fetches data of a validated receipt from AFIP's servers.
        If the receipt exists a ``ReceiptValidation`` instance is
        created and returned, otherwise, returns ``None``.
        If there is already a ``ReceiptValidation`` for this instance,
        returns ``self.validation``.
        This should be used for verification purpose, here's a list of
        some use cases:
         - Incomplete validation process
         - Fetch CAE data from AFIP's servers
        """
        # This may avoid unnecessary revalidation
        if self.is_validated:
            return self.validation

        receipt_data = Receipt.objects.fetch_receipt_data(
            self.receipt_type.code, self.receipt_number, self.point_of_sales
        )

        if not receipt_data:
            return None

        if receipt_data.Resultado == ReceiptValidation.RESULT_APPROVED:
            validation = ReceiptValidation.objects.create(
                result=receipt_data.Resultado,
                cae=receipt_data.CodAutorizacion,
                cae_expiration=parsers.parse_date(receipt_data.FchVto),
                receipt=self,
                processed_date=parsers.parse_datetime(
                    receipt_data.FchProceso,
                ),
            )
            if receipt_data.Observaciones:
                for obs in receipt_data.Observaciones.Obs:
                    observation = Observation.objects.get_or_create(
                        code=obs.Code,
                        message=obs.Msg,
                    )
                validation.observations.add(observation)
            return validation
        return None

    def __repr__(self) -> str:
        return "<Receipt {}: {} {} for {}>".format(
            self.pk,
            self.receipt_type,
            self.receipt_number,
            self.point_of_sales.owner,
        )

    def __str__(self) -> str:
        if self.receipt_number:
            return f"{self.receipt_type} {self.formatted_number}"
        else:
            return _("Unnumbered %s") % self.receipt_type

    class Meta:
        ordering = ("issued_date",)
        verbose_name = _("receipt")
        verbose_name_plural = _("receipts")
        unique_together = [["point_of_sales", "receipt_type", "receipt_number"]]
        # TODO: index_together...


class ReceiptPDFManager(models.Manager):
    def create_for_receipt(self, receipt: Receipt, **kwargs) -> ReceiptPDF:
        """Creates a ReceiptPDF object for a given receipt.

        Does not actually generate the related PDF file.

        All attributes will be completed with the information for the relevant
        :class:`~.PointOfSales` instance.

        :param Receipt receipt: The receipt for the PDF which will be
            generated.
        """
        pdf = ReceiptPDF.objects.create(
            receipt=receipt,
            issuing_name=receipt.point_of_sales.issuing_name,
            issuing_address=receipt.point_of_sales.issuing_address,
            issuing_email=receipt.point_of_sales.issuing_email,
            vat_condition=receipt.point_of_sales.vat_condition,
            gross_income_condition=receipt.point_of_sales.gross_income_condition,
            sales_terms=receipt.point_of_sales.sales_terms,
            **kwargs,
        )
        return pdf


class ReceiptPDF(models.Model):
    """Printable version of a receipt.

    Contains all print-related data of a receipt.

    All ``issuing_*`` fields contain data for the entity issuing the Receipt
    (these may change from one receipt to the next if, for example, the entity
    moved).

    The PDF file itself is saved into the ``pdf_file`` attribute, and is
    generated when saving the model for the first time. If any attributes are
    changed, you should manually call :meth:`~.ReceiptPDF.save_pdf` to
    regenerate the PDF file.

    PDF generation is skipped if the receipt has not been validated.
    """

    def upload_to(self, filename="untitled", instance: ReceiptPDF = None) -> str:
        """
        Returns the full path for generated receipts.

        These are bucketed inside nested directories, to avoid hundreds of
        thousands of children in single directories (which can make reading
        them excessively slow).
        """
        _, extension = os.path.splitext(os.path.basename(filename))
        uuid = uuid4().hex
        buckets = uuid[0:2], uuid[2:4]
        filename = "".join([uuid, extension])

        return os.path.join("afip/receipts", buckets[0], buckets[1], filename)

    receipt = models.OneToOneField(
        Receipt,
        verbose_name=_("receipt"),
        on_delete=models.PROTECT,
    )
    pdf_file = models.FileField(
        verbose_name=_("pdf file"),
        upload_to=upload_to,
        storage=_get_storage_from_settings("AFIP_PDF_STORAGE"),
        blank=True,
        null=True,
        help_text=_("The actual file which contains the PDF data."),
    )
    issuing_name = models.CharField(
        max_length=128,
        verbose_name=_("issuing name"),
    )
    issuing_address = models.TextField(
        _("issuing address"),
    )
    issuing_email = models.CharField(
        max_length=128,
        verbose_name=_("issuing email"),
        blank=True,
        null=True,
    )
    vat_condition = models.CharField(
        max_length=48,
        choices=(
            (
                condition,
                condition,
            )
            for condition in VAT_CONDITIONS
        ),
        verbose_name=_("vat condition"),
    )
    gross_income_condition = models.CharField(
        max_length=48,
        verbose_name=_("gross income condition"),
    )
    client_name = models.CharField(
        max_length=128,
        verbose_name=_("client name"),
    )
    client_address = models.TextField(
        _("client address"),
        blank=True,
    )
    client_vat_condition = models.CharField(
        max_length=48,
        choices=(
            (
                cond,
                cond,
            )
            for cond in CLIENT_VAT_CONDITIONS
        ),
        verbose_name=_("client vat condition"),
    )
    sales_terms = models.CharField(
        max_length=48,
        verbose_name=_("sales terms"),
        help_text=_('Should be something like "Cash", "Payable in 30 days", etc.'),
    )

    objects = ReceiptPDFManager()

    def save_pdf(self, save_model: bool = True) -> None:
        """
        Save the receipt as a PDF related to this model.

        The related :class:`~.Receipt` should be validated first, of course.
        This model instance must have been saved prior to calling this method.

        :param save_model: If True, immediately save this model instance.
        """
        from django_afip.views import ReceiptPDFView

        if not self.receipt.is_validated:
            raise exceptions.DjangoAfipException(
                _("Cannot generate pdf for non-authorized receipt")
            )

        self.pdf_file = File(BytesIO(), name=f"{uuid4().hex}.pdf")
        render_pdf(
            template=ReceiptPDFView().get_template_names(self.receipt),
            file_=self.pdf_file,
            context=ReceiptPDFView.get_context_for_pk(self.receipt_id),
        )

        if save_model:
            self.save()

    def __str__(self) -> str:
        return _("Receipt PDF for %s") % self.receipt_id

    class Meta:
        verbose_name = _("receipt pdf")
        verbose_name_plural = _("receipt pdfs")


class ReceiptEntry(models.Model):
    """An entry in a receipt.

    Each ReceiptEntry represents a line in printable version of a Receipt. You
    should generally have one instance per product or service.

    Note that each entry has a :class:`~.Vat` because a single Receipt can have
    multiple products with different :class:`~.VatType`.
    """

    receipt = models.ForeignKey(
        Receipt,
        related_name="entries",
        verbose_name=_("receipt"),
        on_delete=models.PROTECT,
    )
    description = models.CharField(
        max_length=128,
        verbose_name=_("description"),
    )
    quantity = models.PositiveSmallIntegerField(
        _("quantity"),
    )
    unit_price = models.DecimalField(
        _("unit price"),
        max_digits=15,
        decimal_places=2,
        help_text=_("Price per unit before vat or taxes."),
    )
    vat = models.ForeignKey(
        VatType,
        related_name="receipt_entries",
        verbose_name=_("vat"),
        blank=True,
        null=True,
        on_delete=models.PROTECT,
    )

    @property
    def total_price(self) -> int:
        """The total price for this line (quantity * price)."""
        return self.quantity * self.unit_price

    class Meta:
        verbose_name = _("receipt entry")
        verbose_name_plural = _("receipt entries")


class Tax(models.Model):
    """A tax (type+amount) for a specific Receipt."""

    tax_type = models.ForeignKey(
        TaxType,
        verbose_name=_("tax type"),
        on_delete=models.PROTECT,
    )
    description = models.CharField(
        _("description"),
        max_length=80,
    )
    base_amount = models.DecimalField(
        _("base amount"),
        max_digits=15,
        decimal_places=2,
    )
    aliquot = models.DecimalField(
        _("aliquot"),
        max_digits=5,
        decimal_places=2,
    )
    amount = models.DecimalField(
        _("amount"),
        max_digits=15,
        decimal_places=2,
    )

    receipt = models.ForeignKey(
        Receipt,
        related_name="taxes",
        on_delete=models.PROTECT,
    )

    def compute_amount(self) -> int:
        """Auto-assign and return the total amount for this tax."""
        self.amount = self.base_amount * self.aliquot / 100
        return self.amount

    class Meta:
        verbose_name = _("tax")
        verbose_name_plural = _("taxes")


class Vat(models.Model):
    """A VAT (type+amount) for a specific Receipt."""

    vat_type = models.ForeignKey(
        VatType,
        verbose_name=_("vat type"),
        on_delete=models.PROTECT,
    )
    base_amount = models.DecimalField(
        _("base amount"),
        max_digits=15,
        decimal_places=2,
    )
    amount = models.DecimalField(
        _("amount"),
        max_digits=15,
        decimal_places=2,
    )

    receipt = models.ForeignKey(
        Receipt,
        related_name="vat",
        on_delete=models.PROTECT,
    )

    class Meta:
        verbose_name = _("vat")
        verbose_name_plural = _("vat")


class Observation(models.Model):
    """An observation returned by AFIP.

    AFIP seems to assign re-used codes to Observation, so we actually store
    them as separate objects, and link to them from failed validations.
    """

    code = models.PositiveSmallIntegerField(
        _("code"),
    )
    message = models.CharField(
        _("message"),
        max_length=255,
    )

    class Meta:
        verbose_name = _("observation")
        verbose_name_plural = _("observations")


class ReceiptValidation(models.Model):
    """The validation for a single :class:`~.Receipt`.

    This contains all validation-related data for a receipt, including its CAE
    and the CAE expiration, unless validation has failed.

    The ``observation`` field may contain any data returned by AFIP regarding
    validation failure.
    """

    RESULT_APPROVED = "A"
    RESULT_REJECTED = "R"

    # TODO: replace this with a `successful` boolean field.
    result = models.CharField(
        _("result"),
        max_length=1,
        choices=(
            (RESULT_APPROVED, _("approved")),
            (RESULT_REJECTED, _("rejected")),
        ),
        help_text=_("Indicates whether the validation was succesful or not."),
    )
    processed_date = models.DateTimeField(
        _("processed date"),
    )
    cae = models.CharField(
        _("cae"),
        max_length=14,
        help_text=_("The CAE as returned by the AFIP."),
    )
    cae_expiration = models.DateField(
        _("cae expiration"),
        help_text=_("The CAE expiration as returned by the AFIP."),
    )
    observations = models.ManyToManyField(
        Observation,
        verbose_name=_("observations"),
        related_name="validations",
        help_text=_(
            "The observations as returned by the AFIP. These are generally "
            "present for failed validations."
        ),
    )

    receipt = models.OneToOneField(
        Receipt,
        related_name="validation",
        verbose_name=_("receipt"),
        help_text=_("The Receipt for which this validation applies."),
        on_delete=models.PROTECT,
    )

    def __str__(self) -> str:
        return _("Validation for %s. Result: %s") % (
            self.receipt,
            self.get_result_display(),
        )

    def __repr__(self) -> str:
        return "<{} {}: {} for Receipt {}>".format(
            self.__class__.__name__,
            self.pk,
            self.result,
            self.receipt_id,
        )

    class Meta:
        verbose_name = _("receipt validation")
        verbose_name_plural = _("receipt validations")



from django.db import migrations
from django.db import models


def merge_taxpayer_profile(apps, schema_editor):
    TaxPayerProfile = apps.get_model("afip", "TaxPayerProfile")

    for profile in TaxPayerProfile.objects.all():  # pragma: no cover
        for pos in profile.taxpayer.points_of_sales.all():
            pos.issuing_name = profile.issuing_name
            pos.issuing_address = profile.issuing_address
            pos.issuing_email = profile.issuing_email
            pos.vat_condition = profile.vat_condition
            pos.gross_income_condition = profile.gross_income_condition
            pos.sales_terms = profile.sales_terms
            pos.save()


class Migration(migrations.Migration):

    dependencies = [
        ("afip", "0007_auto_20210409_1641"),
    ]

    operations = [
        migrations.AddField(
            model_name="pointofsales",
            name="gross_income_condition",
            field=models.CharField(
                max_length=48, null=True, verbose_name="gross income condition"
            ),
        ),
        migrations.AddField(
            model_name="pointofsales",
            name="issuing_address",
            field=models.TextField(
                help_text="The address of the issuing entity as shown on receipts.",
                null=True,
                verbose_name="issuing address",
            ),
        ),
        migrations.AddField(
            model_name="pointofsales",
            name="issuing_email",
            field=models.CharField(
                blank=True,
                help_text="The email of the issuing entity as shown on receipts.",
                max_length=128,
                null=True,
                verbose_name="issuing email",
            ),
        ),
        migrations.AddField(
            model_name="pointofsales",
            name="issuing_name",
            field=models.CharField(
                help_text="The name of the issuing entity as shown on receipts.",
                max_length=128,
                null=True,
                verbose_name="issuing name",
            ),
        ),
        migrations.AddField(
            model_name="pointofsales",
            name="sales_terms",
            field=models.CharField(
                help_text=(
                    "The terms of the sale printed onto receipts by default "
                    "(eg: single payment, checking account, etc)."
                ),
                max_length=48,
                null=True,
                verbose_name="sales terms",
            ),
        ),
        migrations.AddField(
            model_name="pointofsales",
            name="vat_condition",
            field=models.CharField(
                choices=[
                    ("IVA Responsable Inscripto", "IVA Responsable Inscripto"),
                    ("IVA Responsable No Inscripto", "IVA Responsable No Inscripto"),
                    ("IVA Exento", "IVA Exento"),
                    ("No Responsable IVA", "No Responsable IVA"),
                    ("Responsable Monotributo", "Responsable Monotributo"),
                ],
                max_length=48,
                null=True,
                verbose_name="vat condition",
            ),
        ),
        migrations.RunPython(merge_taxpayer_profile),
        migrations.DeleteModel(
            name="TaxPayerProfile",
        ),
    ]



from django.conf import settings
from django.contrib import admin
from django.urls import path
from django.views.static import serve

from django_afip import views
from testapp.testmain import views as test_views

urlpatterns = [
    path("admin/", admin.site.urls),
    path(
        "invoices/pdf/<int:pk>",
        views.ReceiptPDFView.as_view(),
        name="receipt_displaypdf_view",
    ),
    path(
        "invoices/pdf/<int:pk>",
        test_views.ReceiptPDFDownloadView.as_view(),
        name="receipt_pdf_view",
    ),
    path(
        "media/<path>",
        serve,
        {"document_root": settings.MEDIA_ROOT},
    ),
]



from pathlib import Path

import pytest

from django_afip import crypto


@pytest.fixture
def signed_data():
    path = Path(__file__).parent.parent.parent.parent / "signed_data.bin"
    with open(path, "rb") as data:
        return data.read()


def test_pkcs7_signing(expired_key: bytes, expired_crt: bytes, signed_data: bytes):
    # Use an expired cert here since this won't change on a yearly basis.
    data = b"Some data."

    actual_data = crypto.create_embeded_pkcs7_signature(data, expired_crt, expired_key)

    # Data after this index DOES vary depending on current time and other settings:
    assert actual_data[:1100] == signed_data[:1100]
    assert len(actual_data) == len(signed_data)



"""
Automatically list fields from django models.

Based on https://djangosnippets.org/snippets/2533/
"""
import inspect

from django.db import models
from django.utils.encoding import force_text
from django.utils.html import strip_tags


def process_docstring(app, what, name, obj, options, lines):
    """
    Process docstrings for django models.

    Process docstrings for django models and add field description from
    their help_text attribute.
    """
    # Only look at objects that inherit from Django's base model class
    if inspect.isclass(obj) and issubclass(obj, models.Model):
        # Grab the field list from the meta class
        fields = obj._meta.get_fields()

        for field in fields:
            if field.name == "id":
                continue

            if not hasattr(field, "help_text") and not hasattr(field, "verbose"):
                # XXX: log these?
                continue

            help_text = strip_tags(force_text(field.help_text))
            verbose_name = force_text(field.verbose_name).capitalize()

            if help_text:
                # Add the model field to the end of the docstring as a param
                # using the help text as the description
                lines.append(f":param {field.attname}: {help_text}")
            else:
                # Add the model field to the end of the docstring as a param
                # using the verbose name as the description
                lines.append(f":param {field.attname}: {verbose_name}")

            # Add the field's type to the docstring
            if isinstance(field, models.ForeignKey):
                to = field.remote_field.model
                lines.append(
                    ":type {}: {} to :class:`~{}.{}`".format(
                        field.attname, type(field).__name__, to.__module__, to.__name__
                    )
                )
            else:
                lines.append(f":type {field.attname}: {type(field).__name__}")

    # Return the extended docstring
    return lines


def setup(app):
    """Register the docstring processor with sphinx."""
    app.connect("autodoc-process-docstring", process_docstring)



from unittest.mock import MagicMock
from unittest.mock import call
from unittest.mock import patch

import pytest
from pytest_django.asserts import assertQuerysetEqual

from django_afip import exceptions
from django_afip import factories
from django_afip import models
from django_afip.factories import ReceiptFactory
from django_afip.factories import ReceiptValidationFactory
from django_afip.factories import ReceiptWithApprovedValidation
from django_afip.factories import ReceiptWithInconsistentVatAndTaxFactory
from django_afip.factories import ReceiptWithVatAndTaxFactory


def test_default_receipt_queryset():
    assert isinstance(models.Receipt.objects.all(), models.ReceiptQuerySet)


@pytest.mark.django_db
def test_validate():
    receipt = ReceiptFactory()
    queryset = models.Receipt.objects.filter(pk=receipt.pk)
    ticket = MagicMock()

    with patch(
        "django_afip.models.ReceiptQuerySet._assign_numbers",
        spec=True,
    ) as mocked_assign_numbers, patch(
        "django_afip.models.ReceiptQuerySet._validate",
        spec=True,
    ) as mocked__validate:
        queryset.validate(ticket)

    assert mocked_assign_numbers.call_count == 1
    assert mocked__validate.call_count == 1
    assert mocked__validate.call_args == call(ticket)


# TODO: Also another tests that checks that we only pass filtered-out receipts.


def test_default_receipt_manager():
    assert isinstance(models.Receipt.objects, models.ReceiptManager)


@pytest.mark.django_db
def test_validate_receipt():
    receipt = ReceiptFactory()
    ticket = MagicMock()
    ticket._called = False

    def fake_validate(qs, ticket=None):
        assertQuerysetEqual(qs, [receipt.pk], lambda r: r.pk)
        ticket._called = True

    with patch(
        "django_afip.models.ReceiptQuerySet.validate",
        fake_validate,
    ):
        receipt.validate(ticket)

    assert ticket._called is True


@pytest.mark.django_db
@pytest.mark.live
def test_validate_invoice(populated_db):
    """Test validating valid receipts."""

    receipt = ReceiptWithVatAndTaxFactory()
    errs = receipt.validate()

    assert len(errs) == 0
    assert receipt.validation.result == models.ReceiptValidation.RESULT_APPROVED
    assert models.ReceiptValidation.objects.count() == 1


@pytest.mark.django_db
@pytest.mark.live
def test_validate_credit_note(populated_db):
    """Test validating valid receipts."""

    # Create a receipt (this credit note relates to it):
    receipt = ReceiptWithVatAndTaxFactory()
    errs = receipt.validate()
    assert len(errs) == 0

    # Create a credit note for the above receipt:
    credit_note = ReceiptWithVatAndTaxFactory(receipt_type__code=8)  # Nota de Crédito B
    credit_note.related_receipts.add(receipt)
    credit_note.save()

    credit_note.validate(raise_=True)
    assert credit_note.receipt_number is not None


@pytest.mark.django_db
@pytest.mark.live
def test_failed_validation(populated_db):
    """Test validating valid receipts."""
    receipt = ReceiptWithInconsistentVatAndTaxFactory()

    errs = receipt.validate()

    assert len(errs) == 1
    # FIXME: We're not creating rejection entries
    # assert receipt.validation.result == models.ReceiptValidation.RESULT_REJECTED
    assert models.ReceiptValidation.objects.count() == 0


@pytest.mark.django_db
@pytest.mark.live
def test_raising_failed_validation(populated_db):
    """Test validating valid receipts."""
    receipt = ReceiptWithInconsistentVatAndTaxFactory()

    with pytest.raises(
        exceptions.ValidationError,
        # Note: AFIP apparently edited this message and added a typo:
        match="DocNro 203012345 no se encuentra registrado en los padrones",
    ):
        receipt.validate(raise_=True)

    # FIXME: We're not creating rejection entries
    # assert receipt.validation.result == models.ReceiptValidation.RESULT_REJECTED
    assert models.ReceiptValidation.objects.count() == 0


@pytest.mark.django_db
@pytest.mark.live
def test_fetch_existing_data(populated_db):
    pos = models.PointOfSales.objects.first()
    rt = models.ReceiptType.objects.get(code=6)
    # last receipt number is needed for testing, it seems they flush old receipts
    # so we can't use a fixed receipt number
    last_receipt_number = models.Receipt.objects.fetch_last_receipt_number(pos, rt)
    receipt = models.Receipt.objects.fetch_receipt_data(
        receipt_type=6,
        receipt_number=last_receipt_number,
        point_of_sales=pos,
    )

    assert receipt.CbteDesde == last_receipt_number
    assert receipt.PtoVta == pos.number


@pytest.mark.django_db
@pytest.mark.live
def test_revalidation_valid_receipt(populated_db):
    """Test revalidation process of a valid receipt."""
    receipt = factories.ReceiptWithVatAndTaxFactory()
    receipt.validate()
    receipt.refresh_from_db()

    old_cae = receipt.validation.cae
    old_validation_pk = receipt.validation.id

    receipt.validation.delete()

    receipt.refresh_from_db()
    assert not receipt.is_validated

    validation = receipt.revalidate()

    assert validation is not None
    assert validation.receipt == receipt
    assert old_cae == validation.cae
    assert old_validation_pk != validation.id


@pytest.mark.live
def test_revalidation_invalid_receipt(populated_db):
    """Test revalidation process of an invalid receipt. (Unexistent receipt)"""
    receipt = factories.ReceiptWithVatAndTaxFactory()
    next_num = (
        models.Receipt.objects.fetch_last_receipt_number(
            receipt.point_of_sales,
            receipt.receipt_type,
        )
        + 1
    )

    receipt.receipt_number = next_num
    receipt.save()

    receipt.refresh_from_db()

    validation = receipt.revalidate()

    assert validation is None


@pytest.mark.django_db
def test_receipt_revalidate_without_receipt_number():
    """Test revalidation process of an invalid receipt. (Receipt without number)"""
    factories.PointOfSalesFactory()
    receipt = factories.ReceiptWithVatAndTaxFactory()
    receipt.refresh_from_db()

    validation = receipt.revalidate()

    assert validation is None


@pytest.mark.django_db
def test_receipt_is_validated_when_not_validated():
    receipt = ReceiptFactory()
    assert not receipt.is_validated


@pytest.mark.django_db
def test_receipt_is_validated_when_validated():
    receipt = ReceiptWithApprovedValidation()
    assert receipt.is_validated


@pytest.mark.django_db
def test_receipt_is_validted_when_failed_validation():
    # These should never really exist,but oh well:
    receipt = ReceiptFactory(receipt_number=None)
    ReceiptValidationFactory(
        receipt=receipt,
        result=models.ReceiptValidation.RESULT_REJECTED,
    )
    assert not receipt.is_validated

    receipt = ReceiptFactory(receipt_number=1)
    ReceiptValidationFactory(
        receipt=receipt,
        result=models.ReceiptValidation.RESULT_REJECTED,
    )
    assert not receipt.is_validated


@pytest.mark.django_db
def test_default_currency_no_currencies():
    receipt = models.Receipt()
    with pytest.raises(models.CurrencyType.DoesNotExist):
        receipt.currency


@pytest.mark.django_db
def test_default_currency_multieple_currencies():
    c1 = factories.CurrencyTypeFactory(pk=2)
    c2 = factories.CurrencyTypeFactory(pk=1)
    c3 = factories.CurrencyTypeFactory(pk=3)

    receipt = models.Receipt()

    assert receipt.currency != c1
    assert receipt.currency == c2
    assert receipt.currency != c3


@pytest.mark.django_db
def test_total_vat_no_vat():
    receipt = ReceiptFactory()

    assert receipt.total_vat == 0


@pytest.mark.django_db
def test_total_vat_multiple_vats():
    receipt = ReceiptFactory()
    factories.VatFactory(receipt=receipt)
    factories.VatFactory(receipt=receipt)

    assert receipt.total_vat == 42


@pytest.mark.live
def test_revalidation_validated_receipt(populated_db):
    """Test revalidation process of a validated receipt."""
    receipt_validation = factories.ReceiptValidationFactory()

    revalidation = receipt_validation.receipt.revalidate()

    assert revalidation is not None
    assert revalidation == receipt_validation


@pytest.mark.django_db
def test_total_vat_proper_filtering():
    receipt = ReceiptFactory()
    factories.VatFactory(receipt=receipt)
    factories.VatFactory()

    assert receipt.total_vat == 21


@pytest.mark.django_db
def test_total_tax_no_tax():
    receipt = ReceiptFactory()

    assert receipt.total_tax == 0


@pytest.mark.django_db
def test_total_tax_multiple_taxes():
    receipt = ReceiptFactory()
    factories.TaxFactory(receipt=receipt)
    factories.TaxFactory(receipt=receipt)

    assert receipt.total_tax == 18


@pytest.mark.django_db
def test_total_tax_proper_filtering():
    receipt = ReceiptFactory()
    factories.TaxFactory(receipt=receipt)
    factories.TaxFactory()

    assert receipt.total_tax == 9


def test_currenty_type_success():
    currency_type = models.CurrencyType(code="011", description="Pesos Uruguayos")
    assert str(currency_type) == "Pesos Uruguayos (011)"


@pytest.mark.django_db
@pytest.mark.live
def test_populate_method(live_ticket):
    assert models.CurrencyType.objects.count() == 0
    models.CurrencyType.objects.populate()
    assert models.CurrencyType.objects.count() == 50



"""Tests for AFIP-WS related classes."""

from datetime import date
from datetime import timedelta
from unittest.mock import patch

import pytest
from django.core import management
from factory.django import FileField
from pytest_django.asserts import assertQuerysetEqual

from django_afip import exceptions
from django_afip import factories
from django_afip import models


@pytest.mark.live
@pytest.mark.django_db
def test_authentication_with_bad_cuit():
    """Test using the wrong cuit for a key pair."""

    taxpayer = factories.AlternateTaxpayerFactory(cuit=20329642339)
    taxpayer.create_ticket("wsfe")

    with pytest.raises(
        exceptions.AfipException,
        # Note: AFIP apparently edited this message and added a typo:
        match="ValidacionDeToken: No apareci[oó] CUIT en lista de relaciones:",
    ):
        taxpayer.fetch_points_of_sales()


@pytest.mark.live
@pytest.mark.django_db
def test_authentication_with_bogus_certificate_exception():
    """Test that using a junk ceritificates raises as expected."""

    # New TaxPayers will fail to save with an invalid cert, but many
    # systems may have very old TaxPayers, externally created, or other
    # stuff, so this scenario might still be possible.
    with patch(
        "django_afip.models.TaxPayer.get_certificate_expiration",
        spec=True,
        return_value=None,
    ):
        taxpayer = factories.TaxPayerFactory(
            key=FileField(data=b"Blah"),
            certificate=FileField(data=b"Blah"),
        )

    with pytest.raises(exceptions.CorruptCertificate) as e:
        taxpayer.create_ticket("wsfe")

    assert not isinstance(e, exceptions.AfipException)


@pytest.mark.live
@pytest.mark.django_db
def test_authentication_with_no_active_taxpayer():
    """Test that no TaxPayers raises an understandable error."""
    with pytest.raises(
        exceptions.AuthenticationError,
        match="There are no taxpayers to generate a ticket.",
    ):
        models.AuthTicket.objects.get_any_active("wsfe")


@pytest.mark.live
@pytest.mark.django_db
def test_authentication_with_expired_certificate_exception():
    """Test that using an expired ceritificate raises as expected."""
    taxpayer = factories.TaxPayerFactory(
        key=FileField(from_path=factories.get_test_file("test_expired.key")),
        certificate=FileField(from_path=factories.get_test_file("test_expired.crt")),
    )

    with pytest.raises(exceptions.CertificateExpired):
        taxpayer.create_ticket("wsfe")


@pytest.mark.live
@pytest.mark.django_db
def test_authentication_with_untrusted_certificate_exception():
    """
    Test that using an untrusted ceritificate raises as expected.
    """
    # Note that we hit production with a sandbox cert here:
    taxpayer = factories.TaxPayerFactory(is_sandboxed=False)

    with pytest.raises(exceptions.UntrustedCertificate):
        taxpayer.create_ticket("wsfe")


@pytest.mark.django_db
def test_population_command():
    """Test the afipmetadata command."""
    management.call_command("afipmetadata")

    assert models.ReceiptType.objects.count() > 0
    assert models.ConceptType.objects.count() > 0
    assert models.DocumentType.objects.count() > 0
    assert models.VatType.objects.count() > 0
    assert models.TaxType.objects.count() > 0
    assert models.CurrencyType.objects.count() > 0


@pytest.mark.django_db
def test_metadata_deserialization():
    """Test that we deserialize descriptions properly."""
    management.call_command("afipmetadata")

    # This assertion is tied to current data, but it validates that we
    # don't mess up encoding/decoding the value we get.
    # It _WILL_ need updating if the upstream value ever changes.
    fac_c = models.ReceiptType.objects.get(code=11)
    assert fac_c.description == "Factura C"


@pytest.mark.django_db
@pytest.mark.live
def test_taxpayer_fetch_points_of_sale(populated_db):
    """Test the ``fetch_points_of_sales`` method."""
    taxpayer = models.TaxPayer.objects.first()
    taxpayer.fetch_points_of_sales()

    points_of_sales = models.PointOfSales.objects.count()
    assert points_of_sales > 0


@pytest.mark.django_db
@pytest.mark.live
def test_receipt_queryset_validate_empty(populated_db):
    factories.ReceiptFactory()

    errs = models.Receipt.objects.none().validate()

    assert errs == []
    assert models.ReceiptValidation.objects.count() == 0


@pytest.mark.django_db
@pytest.mark.live
def test_receipt_queryset_validation_good(populated_db):
    """Test validating valid receipts."""
    r1 = factories.ReceiptWithVatAndTaxFactory()
    r2 = factories.ReceiptWithVatAndTaxFactory()
    r3 = factories.ReceiptWithVatAndTaxFactory()

    errs = models.Receipt.objects.all().validate()

    assert len(errs) == 0
    assert r1.validation.result == models.ReceiptValidation.RESULT_APPROVED
    assert r2.validation.result == models.ReceiptValidation.RESULT_APPROVED
    assert r3.validation.result == models.ReceiptValidation.RESULT_APPROVED
    assert models.ReceiptValidation.objects.count() == 3


@pytest.mark.django_db
@pytest.mark.live
def test_receipt_queryset_validation_bad(populated_db):
    """Test validating invalid receipts."""
    factories.ReceiptWithInconsistentVatAndTaxFactory()
    factories.ReceiptWithInconsistentVatAndTaxFactory()
    factories.ReceiptWithInconsistentVatAndTaxFactory()

    errs = models.Receipt.objects.all().validate()

    assert len(errs) == 1
    assert errs[0] == (
        "Error 10015: Factura B (CbteDesde igual a CbteHasta), DocTipo: "
        "80, DocNro 203012345 no se encuentra registrado en los padrones "
        "de AFIP y no corresponde a una cuit pais."
    )

    assert models.ReceiptValidation.objects.count() == 0


@pytest.mark.django_db
@pytest.mark.live
def test_receipt_queryset_validation_mixed(populated_db):
    """
    Test validating a mixture of valid and invalid receipts.

    Receipts are validated by AFIP in-order, so all receipts previous to
    the bad one are validated, and nothing else is even parsed after the
    invalid one.
    """
    r1 = factories.ReceiptWithVatAndTaxFactory()
    factories.ReceiptWithInconsistentVatAndTaxFactory()
    factories.ReceiptWithVatAndTaxFactory()

    errs = models.Receipt.objects.all().validate()

    assert len(errs) == 1
    assert errs[0] == (
        "Error 10015: Factura B (CbteDesde igual a CbteHasta), DocTipo: "
        "80, DocNro 203012345 no se encuentra registrado en los padrones "
        "de AFIP y no corresponde a una cuit pais."
    )

    assertQuerysetEqual(
        models.ReceiptValidation.objects.all(),
        [r1.pk],
        lambda rv: rv.receipt_id,
    )


@pytest.mark.django_db
@pytest.mark.live
def test_receipt_queryset_validation_validated(populated_db):
    """Test validating invalid receipts."""
    factories.ReceiptWithApprovedValidation()

    errs = models.Receipt.objects.all().validate()

    assert models.ReceiptValidation.objects.count() == 1
    assert errs == []


@pytest.mark.django_db
@pytest.mark.live
def test_receipt_queryset_validation_good_service(populated_db):
    """Test validating a receipt for a service (rather than product)."""
    receipt = factories.ReceiptWithVatAndTaxFactory(
        concept__code=2,
        service_start=date.today() - timedelta(days=10),
        service_end=date.today(),
        expiration_date=date.today() + timedelta(days=10),
    )

    errs = models.Receipt.objects.all().validate()

    assert len(errs) == 0
    assert receipt.validation.result == models.ReceiptValidation.RESULT_APPROVED
    assert models.ReceiptValidation.objects.count() == 1


@pytest.mark.django_db
@pytest.mark.live
def test_receipt_queryset_validation_good_without_tax(populated_db):
    """Test validating valid receipts."""
    receipt = factories.ReceiptFactory(
        point_of_sales=models.PointOfSales.objects.first(),
        total_amount=121,
    )
    factories.VatFactory(vat_type__code=5, receipt=receipt)

    errs = models.Receipt.objects.all().validate()

    assert len(errs) == 0
    assert receipt.validation.result == models.ReceiptValidation.RESULT_APPROVED
    assert models.ReceiptValidation.objects.count() == 1


@pytest.mark.django_db
@pytest.mark.live
def test_receipt_queryset_validation_good_without_vat(populated_db):
    """Test validating valid receipts."""
    receipt = factories.ReceiptFactory(
        point_of_sales=models.PointOfSales.objects.first(),
        receipt_type__code=11,
        total_amount=109,
    )
    factories.TaxFactory(tax_type__code=3, receipt=receipt)

    errs = models.Receipt.objects.all().validate()

    assert len(errs) == 0
    assert receipt.validation.result == models.ReceiptValidation.RESULT_APPROVED
    assert models.ReceiptValidation.objects.count() == 1


@pytest.mark.xfail(reason="Currently not working -- needs to get looked at.")
@pytest.mark.django_db
@pytest.mark.live
def test_receipt_queryset_validation_with_observations(populated_db):
    receipt = factories.ReceiptFactory(
        document_number=20291144404,
        document_type__code=80,
        point_of_sales=models.PointOfSales.objects.first(),
        receipt_type__code=1,
    )
    factories.VatFactory(vat_type__code=5, receipt=receipt)
    factories.TaxFactory(tax_type__code=3, receipt=receipt)

    errs = models.Receipt.objects.all().validate()

    assert len(errs) == 0
    assert receipt.validation.result == models.ReceiptValidation.RESULT_APPROVED
    assert models.ReceiptValidation.objects.count() == 1
    assert models.Observation.objects.count() == 1
    assert receipt.validation.observations.count() == 1


@pytest.mark.django_db
@pytest.mark.live
def test_receipt_queryset_credit_note(populated_db):
    """Test validating valid a credit note."""
    # Create an invoice (code=6) and validate it...
    invoice = factories.ReceiptWithVatAndTaxFactory()

    errs = models.Receipt.objects.filter(pk=invoice.pk).validate()
    assert len(errs) == 0
    assert models.ReceiptValidation.objects.count() == 1

    # Now create a credit note (code=8) and validate it...
    credit = factories.ReceiptWithVatAndTaxFactory(receipt_type__code=8)
    credit.related_receipts.set([invoice])
    credit.save()

    errs = models.Receipt.objects.filter(pk=credit.pk).validate()
    assert len(errs) == 0
    assert models.ReceiptValidation.objects.count() == 2



class DjangoAfipException(Exception):
    """Superclass for all exceptions explicitly thrown by this app."""


class AfipException(DjangoAfipException):
    """
    Wraps around errors returned by AFIP's WS.
    """

    def __init__(self, response):
        Exception.__init__(
            self,
            "Error {}: {}".format(
                response.Errors.Err[0].Code,
                response.Errors.Err[0].Msg,
            ),
        )


class AuthenticationError(DjangoAfipException):
    """
    Raised when there is a non-specific error during an authentication attempt.
    """

    pass


class CertificateExpired(AuthenticationError):
    """
    Raised when an authentication was attempted with an expired certificate.
    """

    pass


class UntrustedCertificate(AuthenticationError):
    """
    Raise when an untrusted certificate is used in an authentication attempt.
    """

    pass


class CorruptCertificate(AuthenticationError):
    """
    Raised when a corrupt ceritificate file is used in an authenticaiton
    attempt.
    """

    pass


class CannotValidateTogether(DjangoAfipException):
    """
    Raised when attempting to validate receipts that cannot be validated
    together (eg: different receipt types, or different point of sales).
    """


class ValidationError(DjangoAfipException):
    """Raised when a single Receipt failed to validate with AFIP's WS."""



import django.core.files.storage
from django.db import migrations
from django.db import models


def merge_taxpayer_extras(apps, schema_editor):
    TaxPayerExtras = apps.get_model("afip", "TaxPayerExtras")

    for extras in TaxPayerExtras.objects.all():  # pragma: no cover
        extras.taxpayer.logo = extras.logo
        extras.taxpayer.save()


class Migration(migrations.Migration):
    dependencies = [
        ("afip", "0004_storages_and_help_texts"),
    ]

    operations = [
        migrations.AddField(
            model_name="taxpayer",
            name="logo",
            field=models.ImageField(
                blank=True,
                help_text="A logo to use when generating printable receipts.",
                null=True,
                storage=django.core.files.storage.FileSystemStorage(),
                upload_to="afip/taxpayers/logos/",
                verbose_name="logo",
            ),
        ),
        migrations.RunPython(merge_taxpayer_extras),
    ]



from unittest import mock
from unittest.mock import patch

import pytest
from django.contrib import messages
from django.contrib.admin import site
from django.http import HttpRequest
from django.test import RequestFactory
from django.utils.translation import gettext as _
from factory.django import FileField
from pytest_django.asserts import assertContains
from pytest_django.asserts import assertNotContains

from django_afip import exceptions
from django_afip import factories
from django_afip import models
from django_afip.admin import ReceiptAdmin  # type: ignore
from django_afip.admin import catch_errors  # type: ignore


def test_certificate_expired():
    admin = mock.MagicMock()
    request = HttpRequest()

    with catch_errors(admin, request):
        raise exceptions.CertificateExpired

    assert admin.message_user.call_count == 1
    assert admin.message_user.call_args == mock.call(
        request,
        _("The AFIP Taxpayer certificate has expired."),
        messages.ERROR,
    )


def test_certificate_untrusted_cert():
    admin = mock.MagicMock()
    request = HttpRequest()

    with catch_errors(admin, request):
        raise exceptions.UntrustedCertificate

    assert admin.message_user.call_count == 1
    assert admin.message_user.call_args == mock.call(
        request,
        _("The AFIP Taxpayer certificate is untrusted."),
        messages.ERROR,
    )


def test_certificate_corrupt_cert():
    admin = mock.MagicMock()
    request = HttpRequest()

    with catch_errors(admin, request):
        raise exceptions.CorruptCertificate

    assert admin.message_user.call_count == 1
    assert admin.message_user.call_args == mock.call(
        request,
        _("The AFIP Taxpayer certificate is corrupt."),
        messages.ERROR,
    )


def test_certificate_auth_error():
    admin = mock.MagicMock()
    request = HttpRequest()

    with catch_errors(admin, request):
        raise exceptions.AuthenticationError

    assert admin.message_user.call_count == 1
    assert admin.message_user.call_args == mock.call(
        request,
        _("An unknown authentication error has ocurred: "),
        messages.ERROR,
    )


def test_without_key(admin_client):
    taxpayer = factories.TaxPayerFactory(key=None)

    response = admin_client.post(
        "/admin/afip/taxpayer/",
        data={"_selected_action": [taxpayer.id], "action": "generate_key"},
        follow=True,
    )

    assert response.status_code == 200
    assertContains(response, "Key generated successfully.")

    taxpayer.refresh_from_db()
    assert "-----BEGIN PRIVATE KEY-----" in taxpayer.key.file.read().decode()


def test_with_key(admin_client):
    taxpayer = factories.TaxPayerFactory(key=FileField(data=b"Blah"))

    response = admin_client.post(
        "/admin/afip/taxpayer/",
        data={"_selected_action": [taxpayer.id], "action": "generate_key"},
        follow=True,
    )

    assert response.status_code == 200
    assertContains(
        response,
        "No keys generated; Taxpayers already had keys.",
    )

    taxpayer.refresh_from_db()
    assert "Blah" == taxpayer.key.file.read().decode()


def test_admin_taxpayer_request_generation_with_csr(admin_client):
    taxpayer = factories.TaxPayerFactory(key=None)
    taxpayer.generate_key()

    response = admin_client.post(
        "/admin/afip/taxpayer/",
        data={"_selected_action": [taxpayer.id], "action": "generate_csr"},
        follow=True,
    )

    assert response.status_code == 200
    assert (
        b"Content-Type: application/pkcs10" in response.serialize_headers().splitlines()
    )
    assertContains(response, "-----BEGIN CERTIFICATE REQUEST-----")


def test_admin_taxpayer_request_generation_without_key(admin_client):
    taxpayer = factories.TaxPayerFactory(key=None)
    taxpayer.generate_key()

    response = admin_client.post(
        "/admin/afip/taxpayer/",
        data={"_selected_action": [taxpayer.id], "action": "generate_csr"},
        follow=True,
    )

    assert response.status_code == 200
    assert (
        b"Content-Type: application/pkcs10" in response.serialize_headers().splitlines()
    )
    assertContains(response, "-----BEGIN CERTIFICATE REQUEST-----")


def test_admin_taxpayer_request_generation_multiple_taxpayers(admin_client):
    taxpayer1 = factories.TaxPayerFactory(key__data=b"Blah")
    taxpayer2 = factories.TaxPayerFactory(key__data=b"Blah", cuit="20401231230")

    response = admin_client.post(
        "/admin/afip/taxpayer/",
        data={
            "_selected_action": [taxpayer1.id, taxpayer2.id],
            "action": "generate_csr",
        },
        follow=True,
    )

    assert response.status_code == 200
    assertContains(response, "Can only generate CSR for one taxpayer at a time")


def test_validation_filters(admin_client):
    """Test the admin validation filters.

    This filters receipts by the validation status.
    """
    validated_receipt = factories.ReceiptFactory()
    failed_validation_receipt = factories.ReceiptFactory()
    not_validated_receipt = factories.ReceiptFactory()

    factories.ReceiptValidationFactory(receipt=validated_receipt)
    factories.ReceiptValidationFactory(
        result=models.ReceiptValidation.RESULT_REJECTED,
        receipt=failed_validation_receipt,
    )

    response = admin_client.get("/admin/afip/receipt/?status=validated")

    assertContains(
        response,
        '<input class="action-select" name="_selected_action" value="{}" '
        'type="checkbox">'.format(validated_receipt.pk),
        html=True,
    )
    assertNotContains(
        response,
        '<input class="action-select" name="_selected_action" value="{}" '
        'type="checkbox">'.format(not_validated_receipt.pk),
        html=True,
    )
    assertNotContains(
        response,
        '<input class="action-select" name="_selected_action" value="{}" '
        'type="checkbox">'.format(failed_validation_receipt.pk),
        html=True,
    )

    response = admin_client.get("/admin/afip/receipt/?status=not_validated")
    assertNotContains(
        response,
        '<input class="action-select" name="_selected_action" value="{}" '
        'type="checkbox">'.format(validated_receipt.pk),
        html=True,
    )
    assertContains(
        response,
        '<input class="action-select" name="_selected_action" value="{}" '
        'type="checkbox">'.format(not_validated_receipt.pk),
        html=True,
    )
    assertContains(
        response,
        '<input class="action-select" name="_selected_action" value="{}" '
        'type="checkbox">'.format(failed_validation_receipt.pk),
        html=True,
    )


@pytest.mark.django_db
def test_receipt_admin_get_exclude():
    admin = ReceiptAdmin(models.Receipt, site)
    request = RequestFactory().get("/admin/afip/receipt")
    request.user = factories.UserFactory()

    assert "related_receipts" in admin.get_fields(request)


@pytest.mark.django_db
def test_receipt_pdf_factories_and_files():
    with_file = factories.ReceiptPDFWithFileFactory()
    without_file = factories.ReceiptPDFFactory()

    assert not without_file.pdf_file
    assert with_file.pdf_file


def test_has_file_filter_all(admin_client):
    """Check that the has_file filter applies properly

    In order to confirm that it's working, we check that the link to the
    object's change page is present, since no matter how we reformat the rows,
    this will always be present as long as the object is listed.
    """
    with_file = factories.ReceiptPDFWithFileFactory()
    without_file = factories.ReceiptPDFFactory()

    response = admin_client.get("/admin/afip/receiptpdf/")
    assertContains(response, f"/admin/afip/receiptpdf/{with_file.pk}/change/")
    assertContains(response, f"/admin/afip/receiptpdf/{without_file.pk}/change/")


def test_has_file_filter_with_file(admin_client):
    with_file = factories.ReceiptPDFWithFileFactory()
    without_file = factories.ReceiptPDFFactory()

    response = admin_client.get("/admin/afip/receiptpdf/?has_file=yes")
    assertContains(response, f"/admin/afip/receiptpdf/{with_file.pk}/change/")
    assertNotContains(response, f"/admin/afip/receiptpdf/{without_file.pk}/change/")


def test_has_file_filter_without_file(admin_client):
    with_file = factories.ReceiptPDFWithFileFactory()
    without_file = factories.ReceiptPDFFactory()

    response = admin_client.get("/admin/afip/receiptpdf/?has_file=no")
    assertNotContains(response, f"/admin/afip/receiptpdf/{with_file.pk}/change/")
    assertContains(response, f"/admin/afip/receiptpdf/{without_file.pk}/change/")


def test_validate_certs_action_success(admin_client):
    receipt = factories.ReceiptFactory()

    with patch(
        "django_afip.models.ReceiptQuerySet.validate", spec=True, return_value=[]
    ) as validate:
        response = admin_client.post(
            "/admin/afip/receipt/",
            data={"_selected_action": [receipt.id], "action": "validate"},
            follow=True,
        )

    assert response.status_code == 200
    assert validate.call_count == 1
    assert list(response.context["messages"]) == []


def test_validate_certs_action_errors(admin_client):
    receipt = factories.ReceiptFactory()

    with patch(
        "django_afip.models.ReceiptQuerySet.validate",
        spec=True,
        return_value=["Something went wrong"],
    ) as validate:
        response = admin_client.post(
            "/admin/afip/receipt/",
            data={"_selected_action": [receipt.id], "action": "validate"},
            follow=True,
        )

    assert response.status_code == 200
    assert validate.call_count == 1

    messages = list(response.context["messages"])
    assert len(messages) == 1

    message = messages[0].message
    assert message == "Receipt validation failed: ['Something went wrong']."


def test_admin_fetch_points_of_sales(admin_client):
    taxpayer1 = factories.TaxPayerFactory()
    taxpayer2 = factories.TaxPayerFactory(cuit="20401231230")
    with patch(
        "django_afip.models.TaxPayer.fetch_points_of_sales",
        spec=True,
        return_value=[("dummy-point-of-sales", True), ("another", False)],
    ):
        response = admin_client.post(
            "/admin/afip/taxpayer/",
            data={
                "_selected_action": [taxpayer1.id, taxpayer2.id],
                "action": "fetch_points_of_sales",
            },
            follow=True,
        )

    assert response.status_code == 200

    messages = [msg.message for msg in list(response.context["messages"])]
    assert len(messages) == 2

    assert "2 points of sales already existed." in messages
    assert "2 points of sales created." in messages



from datetime import date

import pytest
from django.test import Client
from django.test import TestCase
from django.urls import reverse
from pytest_django.asserts import assertContains
from pytest_django.asserts import assertHTMLEqual

from django_afip import factories
from django_afip import views


class ReceiptPDFTestCase(TestCase):
    def test_html_view(self):
        """Test the HTML generation view."""
        pdf = factories.ReceiptPDFFactory(
            receipt__concept__code=1,
            receipt__issued_date=date(2017, 5, 15),
            receipt__receipt_type__code=11,
            receipt__point_of_sales__owner__logo=None,
        )
        factories.ReceiptValidationFactory(receipt=pdf.receipt)

        client = Client()
        response = client.get(
            "{}?html=true".format(
                reverse("receipt_displaypdf_view", args=(pdf.receipt.pk,))
            )
        )

        assertHTMLEqual(
            response.content.decode(),
            """
<!DOCTYPE html>
<html>
  <head>
    <meta charset="utf-8">
    <link rel="stylesheet" href="/static/receipts/receipt.css">
  </head>
  <body>
    <div class="receipt">

      <header>
        <div class="taxpayer-details group">
          <address>
            <strong>Alice Doe</strong><br>
            Happy Street 123, CABA<br>

            Responsable Monotributo<br>
          </address>

          <div class="receipt-type">
            <div class="identifier">
              C
            </div>
            <div class="code">
              Código 11
            </div>
          </div>

          <div class="receipt-details">
            <div class="receipt-type-description">
              Factura C
            </div>
            <strong>Nº</strong> None
            <br>
            <strong>Fecha de emisión:</strong> <time>May 15, 2017</time><br>
            <strong>CUIT:</strong> 20-32964233-0<br>
            <strong>Ingresos Brutos:</strong> Convenio Multilateral<br>
            <strong>Inicio de Actividades:</strong> Oct. 3, 2011
          </div>
        </div>

        <hr>

        <div class="client">
          <div><strong>Facturado a:</strong></div>
          <div class="sale-conditions">
            <strong>Condición de IVA:</strong> Consumidor Final<br>
            <strong>Condición de Pago:</strong> Contado
          </div>
          <div class="client-data">
            John Doe, DNI 203012345<br>
            La Rioja 123<br />X5000EVX Córdoba<br>
          </div>
        </div>

      </header>

      <hr>

      <table>
        <thead>
          <tr>
            <th>Descripción</th>
            <th>Cantidad</th>
            <th>Precio Unitario</th>
            <th>Monto</th>
          </tr>
        </thead>
        <tbody>

        </tbody>
        <tfoot>
          <tr>
            <td></td>
            <td></td>
            <td></td>
            <td>130.00</td>
          </tr>
        </tfoot>
      </table>

      <footer>
        <div class="qrcode">
          <img src="data:image/png;base64,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">
        </div>

        <p class="cae">
          <strong>CAE</strong>
          67190616790549
          <strong>Vto CAE</strong>
          July 12, 2017
        </p>

        Consultas de validez:
        <a href="http://www.afip.gob.ar/genericos/consultacae/">
          http://www.afip.gob.ar/genericos/consultacae/
        </a>
        <br>
        Teléfono Gratuito CABA, Área de Defensa y Protección al Consumidor.
        Tel 147
      </footer>

    </div>
  </body>
</html>
            """,  # noqa: E501: It's just long stuff. :(
        )

    def test_logo_in_html(self):
        """Test the HTML generation view."""
        pdf = factories.ReceiptPDFFactory()
        factories.ReceiptValidationFactory(receipt=pdf.receipt)

        client = Client()
        response = client.get(
            "{}?html=true".format(
                reverse("receipt_displaypdf_view", args=(pdf.receipt.pk,))
            )
        )

        assertContains(
            response,
            """
            <address>
            <img src="data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAIAAAACCAIAAAD91JpzAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAB3RJTUUH4QgLERgyoZWu2QAAAB1pVFh0Q29tbWVudAAAAAAAQ3JlYXRlZCB3aXRoIEdJTVBkLmUHAAAAEklEQVQI12P4//8/AwT8//8fACnkBft7DmIIAAAAAElFTkSuQmCC" alt="Logo"><br>
            <strong>Alice Doe</strong><br>
            Happy Street 123, CABA<br>

            Responsable Monotributo<br>
            </address>
            """,  # noqa: E501: It's just long stuff. :(
            html=True,
        )

    def test_pdf_view(self):
        """
        Test the PDF generation view.
        """
        taxpayer = factories.TaxPayerFactory()

        pdf = factories.ReceiptPDFFactory(
            receipt__point_of_sales__owner=taxpayer,
        )
        factories.ReceiptValidationFactory(receipt=pdf.receipt)

        client = Client()
        response = client.get(reverse("receipt_pdf_view", args=(pdf.receipt.pk,)))

        assert response.status_code == 200
        assert response.content[:7] == b"%PDF-1."

        headers = sorted(response.serialize_headers().decode().splitlines())
        assert "Content-Type: application/pdf" in headers


@pytest.mark.django_db
def test_template_discovery(client):
    taxpayer = factories.TaxPayerFactory(cuit="20329642330")
    pdf = factories.ReceiptPDFFactory(
        receipt__point_of_sales__owner=taxpayer,
        receipt__point_of_sales__number=9999,
        receipt__receipt_type__code=6,
    )
    factories.ReceiptValidationFactory(receipt=pdf.receipt)

    client = Client()
    response = client.get(
        "{}?html=true".format(
            reverse("receipt_displaypdf_view", args=(pdf.receipt.pk,))
        )
    )

    assert response.content == b"This is a dummy template to test template discovery.\n"


class ReceiptPDFViewDownloadNameTestCase(TestCase):
    def test_download_name(self):
        factories.ReceiptFactory(pk=9, receipt_number=32)

        view = views.ReceiptPDFView()
        view.kwargs = {"pk": 9}

        assert view.get_download_name() == "0001-00000032.pdf"



from dataclasses import dataclass

from django_afip import clients


@dataclass(frozen=True)
class ServerStatus:
    """A dataclass holding the server's reported status.

    An instance is truthy if all services are okay, or evaluates to ``False``
    if at least one isn't::

        if not server_status:
            print("At least one service is down")
        else
            print("All serivces are up")
    """

    #: Whether the application server is working.
    app: bool
    #: Whether the database server is working.
    db: bool
    #: Whether the authentication server is working.
    auth: bool

    def __bool__(self):
        return self.app and self.db and self.auth


def get_server_status(production: bool) -> ServerStatus:
    """Return the status of AFIP's WS servers

    :param production: Whether to check the production servers. If false, the
        testing servers will be checked instead.
    """
    client = clients.get_client("wsfe", not production)
    response = client.service.FEDummy()

    return ServerStatus(
        app=response["AppServer"] == "OK",
        db=response["DbServer"] == "OK",
        auth=response["AuthServer"] == "OK",
    )



from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("afip", "0005_flatten_taxpayer_extras"),
    ]

    operations = [
        migrations.DeleteModel(
            name="TaxPayerExtras",
        ),
    ]



import django.db.models.deletion
from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ("afip", "0006_delete_taxpayerextras"),
    ]

    operations = [
        migrations.AlterField(
            model_name="authticket",
            name="service",
            field=models.CharField(
                help_text="Service for which this ticket has been authorized.",
                max_length=6,
                verbose_name="service",
            ),
        ),
        migrations.AlterField(
            model_name="receipt",
            name="currency_quote",
            field=models.DecimalField(
                decimal_places=6,
                default=1,
                help_text="The currency's quote on the day this receipt was issued.",
                max_digits=10,
                verbose_name="currency quote",
            ),
        ),
        migrations.AlterField(
            model_name="receipt",
            name="document_number",
            field=models.BigIntegerField(
                help_text="The document number of the recipient of this receipt.",
                verbose_name="document number",
            ),
        ),
        migrations.AlterField(
            model_name="receipt",
            name="document_type",
            field=models.ForeignKey(
                help_text="The document type of the recipient of this receipt.",
                on_delete=django.db.models.deletion.PROTECT,
                related_name="receipts",
                to="afip.documenttype",
                verbose_name="document type",
            ),
        ),
        migrations.AlterField(
            model_name="receipt",
            name="exempt_amount",
            field=models.DecimalField(
                decimal_places=2,
                help_text=(
                    "Only for categories which are tax-exempt.<br>"
                    "For C-type receipts, this must be zero."
                ),
                max_digits=15,
                verbose_name="exempt amount",
            ),
        ),
        migrations.AlterField(
            model_name="receipt",
            name="issued_date",
            field=models.DateField(
                help_text="Can diverge up to 5 days for good, or 10 days otherwise.",
                verbose_name="issued date",
            ),
        ),
        migrations.AlterField(
            model_name="receipt",
            name="net_taxed",
            field=models.DecimalField(
                decimal_places=2,
                help_text=(
                    "The total amount to which taxes apply.<br>"
                    "For C-type receipts, this is equal to the subtotal."
                ),
                max_digits=15,
                verbose_name="total taxable amount",
            ),
        ),
        migrations.AlterField(
            model_name="receipt",
            name="net_untaxed",
            field=models.DecimalField(
                decimal_places=2,
                help_text=(
                    "The total amount to which taxes do not apply.<br>"
                    "For C-type receipts, this must be zero."
                ),
                max_digits=15,
                verbose_name="total untaxable amount",
            ),
        ),
        migrations.AlterField(
            model_name="receiptpdf",
            name="sales_terms",
            field=models.CharField(
                help_text='Should be something like "Cash", "Payable in 30 days", etc.',
                max_length=48,
                verbose_name="sales terms",
            ),
        ),
        migrations.AlterField(
            model_name="receiptvalidation",
            name="cae",
            field=models.CharField(
                help_text="The CAE as returned by the AFIP.",
                max_length=14,
                verbose_name="cae",
            ),
        ),
        migrations.AlterField(
            model_name="receiptvalidation",
            name="cae_expiration",
            field=models.DateField(
                help_text="The CAE expiration as returned by the AFIP.",
                verbose_name="cae expiration",
            ),
        ),
        migrations.AlterField(
            model_name="receiptvalidation",
            name="receipt",
            field=models.OneToOneField(
                help_text="The Receipt for which this validation applies.",
                on_delete=django.db.models.deletion.PROTECT,
                related_name="validation",
                to="afip.receipt",
                verbose_name="receipt",
            ),
        ),
        migrations.AlterField(
            model_name="receiptvalidation",
            name="result",
            field=models.CharField(
                choices=[("A", "approved"), ("R", "rejected")],
                help_text="Indicates whether the validation was succesful or not.",
                max_length=1,
                verbose_name="result",
            ),
        ),
        migrations.AlterField(
            model_name="taxpayer",
            name="certificate_expiration",
            field=models.DateTimeField(
                editable=False,
                help_text=(
                    "Stores expiration for the current certificate.<br>"
                    "Note that this field is updated pre-save, so the value may be "
                    "invalid for unsaved models."
                ),
                null=True,
                verbose_name="certificate expiration",
            ),
        ),
        migrations.AlterField(
            model_name="taxpayer",
            name="is_sandboxed",
            field=models.BooleanField(
                help_text=(
                    "Indicates if this taxpayer should use with the sandbox "
                    "servers rather than the production servers."
                ),
                verbose_name="is sandboxed",
            ),
        ),
        migrations.AlterField(
            model_name="taxpayerprofile",
            name="issuing_address",
            field=models.TextField(
                help_text="The address of the issuing entity as shown on receipts.",
                verbose_name="issuing address",
            ),
        ),
        migrations.AlterField(
            model_name="taxpayerprofile",
            name="issuing_email",
            field=models.CharField(
                blank=True,
                help_text="The email of the issuing entity as shown on receipts.",
                max_length=128,
                null=True,
                verbose_name="issuing email",
            ),
        ),
        migrations.AlterField(
            model_name="taxpayerprofile",
            name="issuing_name",
            field=models.CharField(
                help_text="The name of the issuing entity as shown on receipts.",
                max_length=128,
                verbose_name="issuing name",
            ),
        ),
    ]



from django.db import migrations
from django.db import models

import django_afip.models


class Migration(migrations.Migration):

    dependencies = [
        ("afip", "0003_issuance_type_length"),
    ]

    operations = [
        migrations.AlterField(
            model_name="receiptpdf",
            name="pdf_file",
            field=models.FileField(
                blank=True,
                help_text="The actual file which contains the PDF data.",
                null=True,
                storage=django_afip.models._get_storage_from_settings(
                    "AFIP_PDF_STORAGE"
                ),
                upload_to=django_afip.models.ReceiptPDF.upload_to,
                verbose_name="pdf file",
            ),
        ),
        migrations.AlterField(
            model_name="taxpayer",
            name="certificate",
            field=models.FileField(
                blank=True,
                null=True,
                storage=django_afip.models._get_storage_from_settings(
                    "AFIP_CERT_STORAGE"
                ),
                upload_to="afip/taxpayers/certs/",
                verbose_name="certificate",
            ),
        ),
        migrations.AlterField(
            model_name="taxpayer",
            name="key",
            field=models.FileField(
                blank=True,
                null=True,
                storage=django_afip.models._get_storage_from_settings(
                    "AFIP_KEY_STORAGE"
                ),
                upload_to="afip/taxpayers/keys/",
                verbose_name="key",
            ),
        ),
        migrations.AlterField(
            model_name="taxpayerextras",
            name="logo",
            field=models.ImageField(
                blank=True,
                help_text="A logo to use when generating printable receipts.",
                null=True,
                storage=django_afip.models._get_storage_from_settings(
                    "AFIP_LOGO_STORAGE"
                ),
                upload_to="afip/taxpayers/logos/",
                verbose_name="logo",
            ),
        ),
    ]
